from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import export
from dissect.target.plugins.os.windows.dpapi.keyprovider.defaultpassword.defaultpassword import (
    DefaultPasswordKeyProvider,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

defaultpassword_def = """
struct DefaultPassword {
    DWORD   length;
    char    flags[4*3];
    WCHAR   data[length/2];
    char    checksum_or_guid[0x10];
};
"""

c_defaultpassword = cstruct().load(defaultpassword_def)


class LSADefaultPasswordKeyProviderPlugin(DefaultPasswordKeyProvider):
    """Windows LSA DefaultPassword key provider plugin."""

    __namespace__ = "dpapi.keyprovider.defaultpassword.lsa"

    def check_compatible(self) -> None:
        if not self.target.has_function("lsa"):
            raise UnsupportedPluginError("LSA plugin not available on target")

    @export(output="yield")
    def keys(self) -> Iterator[tuple[str, str]]:
        """Yield Windows LSA DefaultPassword decrypted strings.

        Extracts decrypted ``DefaultPassword`` values from the LSA.

        References:
            - https://learn.microsoft.com/en-us/troubleshoot/windows-server/user-profiles-and-logon/turn-on-automatic-logon
        """

        # Search for DefaultPassword values in the LSA
        for secret in ["DefaultPassword", "DefaultPassword_OldVal"]:
            if default_pass := self.target.lsa._secrets.get(secret):
                try:
                    value = c_defaultpassword.DefaultPassword(default_pass).data
                except Exception as e:
                    self.target.log.warning("Failed to parse LSA %s value (%r): %s", secret, default_pass, e)
                    self.target.log.debug("", exc_info=e)
                    continue
                yield self.__namespace__, value



from __future__ import annotations

import stat
from functools import cached_property
from typing import TYPE_CHECKING, BinaryIO, Callable, TypeVar

from dissect.util.stream import AlignedStream

from dissect.target.exceptions import NotADirectoryError
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil
from dissect.target.helpers.nfs.client.mount import Client as MountClient
from dissect.target.helpers.nfs.client.nfs import Client as NfsClient
from dissect.target.helpers.nfs.nfs3 import (
    EntryPlus,
    FileAttributes,
    FileHandle,
    FileType,
    MountProc,
    NfsProgram,
    NfsVersion,
)
from dissect.target.helpers.sunrpc.client import AuthScheme, LocalPortPolicy, auth_null
from dissect.target.helpers.sunrpc.client import Client as SunRpcClient

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

ConCredentials = TypeVar("ConCredentials")
ConVerifier = TypeVar("ConVerifier")

ClientFactory = Callable[[], NfsClient]

# Set auth scheme programmatically, given a root filehandle and a list of supported auth flavors
# Example: Trying multiple auth schemes until one works
AuthSetter = Callable[[NfsClient, FileHandle, list[int]], None]


class AuthFlavorNotSupported(Exception):
    def __init__(self, supported_flavors: list[int], provided_flavor: int):
        self.supported = supported_flavors
        self.provided = provided_flavor

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__} Auth flavor {self.provided} not supported. Supported flavors: {self.supported}"
        )


class NfsFilesystem(Filesystem):
    """Filesystem implementation of a NFS share

    The connection is lazily established to not waste resources.
    Use the ``connect`` method to conveniently create a new instance.
    """

    __type__ = "nfs"

    def __init__(self, client_factory: ClientFactory, root_handle: FileHandle):
        super().__init__()
        self._client_factory = client_factory
        self._root_handle = root_handle

    @classmethod
    def connect(
        cls,
        address: str,
        exported_dir: str,
        auth: AuthScheme[ConCredentials, ConVerifier] | AuthSetter,
        local_port: int | LocalPortPolicy = 0,
        timeout_in_seconds: float | None = 5.0,
    ) -> Self:
        """Utility function to setup a connection to a NFS share.

        Args:
            hostname: The remote hostname.
            port: The remote port.
            auth: The authentication scheme.
            local_port: The local port to bind to.
                If equal to ``LocalPortPolicy.PRIVILEGED`` or -1, bind to the first free privileged port.
                If equal to ``LocalPortPolicy.ANY`` or 0, bind to any free port.
                Otherwise, bind to the specified port.
            timeout_in_seconds: The timeout for making the connection.
        """
        with SunRpcClient.connect_port_mapper(address, timeout_in_seconds) as port_mapper_client:
            mount_port = port_mapper_client.query_port_mapping(MountProc.program, version=MountProc.version)
            nfs_port = port_mapper_client.query_port_mapping(NfsProgram, version=NfsVersion)

        # We eagerly mount the share because the root file handle is required for succesful mounting
        with MountClient.connect(address, mount_port, local_port) as mount_client:
            mount = mount_client.mount(exported_dir)

        def client_factory() -> NfsClient:
            if isinstance(auth, AuthScheme):
                if auth.flavor not in mount.auth_flavors:
                    raise AuthFlavorNotSupported(mount.auth_flavors, auth.flavor)
                return NfsClient.connect(address, nfs_port, auth, local_port)

            client = NfsClient.connect(address, nfs_port, auth_null(), local_port)
            auth(client, mount.filehandle, mount.auth_flavors)
            return client

        return NfsFilesystem(client_factory, mount.filehandle)

    @cached_property
    def _client(self) -> NfsClient:
        return self._client_factory()

    @staticmethod
    def detect(_: BinaryIO) -> bool:
        raise TypeError("Detect is not allowed on a NfsFilesystem class")  # :/

    def get(self, path: str, relentry: NfsFilesystemEntry | None = None) -> NfsFilesystemEntry:
        """Get a filesystem entry.

        Args:
            path: The path to the entry. The path is relative to ``relentry``, if provided.
            relentry: The relative entry to start from. If not provided, the root entry is used.
        """

        current_handle = relentry.entry if relentry else self._root_handle
        path = fsutil.normalize(path, self.alt_separator).strip("/")

        if not path:
            return NfsFilesystemEntry(self, "/", current_handle)

        for segment in path.split("/"):
            result = self._client.lookup(segment, current_handle)
            current_handle = result.object

        return NfsFilesystemEntry(self, path, result.object, result.obj_attributes)


class NfsFilesystemEntry(FilesystemEntry):
    fs: NfsFilesystem
    entry: FileHandle

    def __init__(self, fs: NfsFilesystem, path: str, file_handle: FileHandle, attributes: FileAttributes | None = None):
        super().__init__(fs, path, file_handle)
        self._backing_attributes = attributes

    @property
    def _attributes(self) -> FileAttributes:
        if self._backing_attributes is None:
            self._backing_attributes = self.fs._client.getattr(self.entry)
        return self._backing_attributes

    def get(self, path: str) -> NfsFilesystemEntry:
        """Get a new filesystem entry relative to this entry"""
        if not self.is_dir():
            raise NotADirectoryError

        return self.fs.get(path, relentry=self)

    def is_file(self, follow_symlinks: bool = True) -> bool:
        # Not using _resolve since it upcasts. ("Self" from Python 3.11 would solve this)
        # Or create a subclass of Filesystementry for symlinks.
        if follow_symlinks and self.is_symlink():
            return self.readlink_ext().is_file()

        return self._attributes.type == FileType.REG

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        if follow_symlinks and self.is_symlink():
            return self.readlink_ext().is_dir()

        return self._attributes.type == FileType.DIR

    def is_symlink(self) -> bool:
        return self._attributes.type == FileType.LNK

    def readlink(self) -> str:
        return self.fs._client.readlink(self.entry)

    def readlink_ext(self) -> NfsFilesystemEntry:
        target = self.fs._client.readlink(self.entry)
        return self.fs.get(target)  # The target is an absolute path

    def _iterdir(self) -> Iterator[EntryPlus]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        yield from self.fs._client.readdir(self.entry).entries

    def iterdir(self) -> Iterator[str]:
        for entry in self._iterdir():
            yield entry.name

    def scandir(self) -> Iterator[FilesystemEntry]:
        for entry in self._iterdir():
            yield NfsFilesystemEntry(self.fs, entry.name, entry.handle, entry.attributes)

    def open(self) -> NfsStream:
        # Pass size if available but don't sweat it
        size = self._backing_attributes.size if self._backing_attributes else None
        return NfsStream(self.fs._client, self.entry, size)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        if follow_symlinks and self.is_symlink():
            return self.readlink_ext().lstat()

        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        attributes = self._attributes

        st_info = fsutil.stat_result(
            [
                attributes.mode | self._mode_file_type(attributes.type),
                fsutil.generate_addr(self.path, alt_separator=self.fs.alt_separator),
                attributes.fsid,
                attributes.nlink,
                attributes.uid,
                attributes.gid,
                attributes.size,
                attributes.atime.seconds,
                attributes.mtime.seconds,
                attributes.ctime.seconds,
            ]
        )

        st_info.st_atime_ns = attributes.atime.nseconds
        st_info.st_mtime_ns = attributes.mtime.nseconds
        st_info.st_ctime_ns = attributes.ctime.nseconds

        return st_info

    def _mode_file_type(self, type: FileType) -> int:
        if type == FileType.DIR:
            return stat.S_IFDIR
        if type == FileType.REG:
            return stat.S_IFREG
        if type == FileType.LNK:
            return stat.S_IFLNK
        return 0o000000


class NfsStream(AlignedStream):
    def __init__(self, client: NfsClient, file_handle: FileHandle, size: int | None):
        super().__init__(size, NfsClient.READ_CHUNK_SIZE)
        self._client = client
        self._file_handle = file_handle

    def _read(self, offset: int, size: int) -> bytes:
        data = self._client.readfile(self._file_handle, offset, size)
        return b"".join(data)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Final

from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import NamespacePlugin

if TYPE_CHECKING:
    from typing_extensions import Self

PackageManagerLogRecord = TargetRecordDescriptor(
    "unix/packagemanager/log",
    [
        ("datetime", "ts"),
        ("string", "package_manager"),
        ("string", "operation"),
        ("string", "package_name"),
        ("string", "command"),
        ("string", "requested_by_user"),
    ],
)


class OperationTypes(Enum):
    """Valid operation types."""

    Install = "install"
    Update = "update"
    Downgrade = "downgrade"
    Remove = "remove"
    Other = "other"

    __MAPPING__: Final[dict[str, list[str]]] = {
        Install: ["install", "installed", "reinstall"],
        Update: ["update", "updated", "upgrade"],
        Downgrade: ["downgrade"],
        Remove: ["remove", "removed", "erased", "purge"],
        Other: ["command"],
    }

    @classmethod
    def infer(cls, keyword: str) -> Self:
        keyword = keyword.strip().lower()
        for key, values in cls.__MAPPING__.items():
            if keyword in values:
                return OperationTypes(key)
        return OperationTypes.Other


class PackageManagerPlugin(NamespacePlugin):
    __namespace__ = "packagemanager"



from __future__ import annotations

import io
from pathlib import Path
from typing import Callable
from unittest.mock import call, patch

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loader import open as loader_open
from dissect.target.loaders.multiraw import MultiRawLoader
from dissect.target.target import Target


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], tmp_path: Path) -> None:
    """Test that we correctly use ``MultiRawLoader`` when opening a ``Target``."""
    file1 = tmp_path / "file1.bin"
    file2 = tmp_path / "file2.bin"
    file1.touch()
    file2.touch()

    fs = VirtualFilesystem()
    fs.map_file_fh("/dir/file1.bin", io.BytesIO())
    fs.map_file_fh("/dir/file2.bin", io.BytesIO())

    for path in (f"{file1}+{file2}", fs.path("/dir/file1.bin+/dir/file2.bin")):
        with patch("dissect.target.container.open"), patch("dissect.target.target.Target.apply"):
            target = opener(path)
            assert isinstance(target._loader, MultiRawLoader)

            if isinstance(path, str):
                assert target.path == Path(path)
            else:
                assert target.path == path


def test_local(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that ``MultiRawLoader`` works with local files, absolute and relative."""
    (tmp_path / "file1.bin").touch()
    (tmp_path / "file2.bin").touch()

    root = str(tmp_path)

    loader = loader_open(Path(f"{root}/file1.bin"))
    assert not isinstance(loader, MultiRawLoader)

    loader = loader_open(Path(f"{root}/file1.bin+{root}/file2.bin"))
    assert isinstance(loader, MultiRawLoader)

    with patch("dissect.target.container.open") as mock_container_open:
        t = Target()
        loader.map(t)

        assert len(t.disks) == 2
        assert mock_container_open.call_count == 2
        mock_container_open.assert_has_calls(
            [
                call(Path(f"{root}/file1.bin")),
                call(Path(f"{root}/file2.bin")),
            ]
        )

    monkeypatch.chdir(root)
    assert Path("file1.bin").exists()

    loader = loader_open(Path("file1.bin"))
    assert not isinstance(loader, MultiRawLoader)

    loader = loader_open(Path("file1.bin+file2.bin"))
    assert isinstance(loader, MultiRawLoader)


def test_path() -> None:
    """Test that ``MultiRawLoader`` works with a ``TargetPath``."""
    fs = VirtualFilesystem()
    fs.map_file_fh("/dir/file1.bin", io.BytesIO())
    fs.map_file_fh("/dir/file2.bin", io.BytesIO())

    loader = loader_open(fs.path("/dir/file1.bin"))
    assert not isinstance(loader, MultiRawLoader)

    loader = loader_open(fs.path("/dir/file1.bin+/dir/file2.bin"))
    assert isinstance(loader, MultiRawLoader)

    with patch("dissect.target.container.open") as mock_container_open:
        t = Target()
        loader.map(t)

        assert len(t.disks) == 2
        assert mock_container_open.call_count == 2
        mock_container_open.assert_has_calls(
            [
                call(fs.path("/dir/file1.bin")),
                call(fs.path("/dir/file2.bin")),
            ]
        )



from __future__ import annotations

import textwrap
from io import BytesIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.os.unix.cronjobs import CronjobPlugin, CronjobRecord

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_unix_cronjobs_system(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we correctly infer the username of the cronjob from the command."""

    fs_unix.map_file_fh(
        "/etc/crontab", BytesIO(b"17 *	* * *	root	cd / && run-parts --report /etc/cron.hourly")
    )
    target_unix_users.add_plugin(CronjobPlugin)

    results = list(target_unix_users.cronjobs())
    assert len(results) == 1
    assert results[0].minute == "17"
    assert results[0].hour == "*"
    assert results[0].day == "*"
    assert results[0].month == "*"
    assert results[0].weekday == "*"
    assert results[0].username == "root"
    assert results[0].command == "cd / && run-parts --report /etc/cron.hourly"


def test_unix_cronjobs_user(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we correctly infer the username of the crontab from the file path."""

    fs_unix.map_file_fh("/var/spool/cron/crontabs/user", BytesIO(b"0 0 * * * /path/to/example.sh\n"))
    target_unix_users.add_plugin(CronjobPlugin)

    results = list(target_unix_users.cronjobs())
    assert len(results) == 1
    assert results[0].minute == "0"
    assert results[0].hour == "0"
    assert results[0].day == "*"
    assert results[0].month == "*"
    assert results[0].weekday == "*"
    assert results[0].username == "user"
    assert results[0].command == "/path/to/example.sh"


def test_unix_cronjobs_env(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we parse environment variables inside crontab files correctly."""

    crontab = """
    FOO=bar
    PATH=/path/to/some/example
    0 0 * * * example.sh
    """

    fs_unix.map_file_fh("/etc/crontab", BytesIO(textwrap.dedent(crontab).encode()))

    results = list(target_unix.cronjobs())
    assert len(results) == 3
    assert results[0].key == "FOO"
    assert results[0].value == "bar"
    assert results[1].key == "PATH"
    assert results[1].value == "/path/to/some/example"
    assert results[2].command == "example.sh"
    assert results[2].username is None


@pytest.mark.parametrize(
    ("cron_line", "expected_output", "expected_user"),
    [
        (
            "0 0 * * * FOO=bar    /path/to/some/script.sh",
            {
                "command": "FOO=bar    /path/to/some/script.sh",
                "day": "*",
                "hour": "0",
                "minute": "0",
                "month": "*",
                "source": "/etc/crontab",
                "weekday": "*",
            },
            None,
        ),
        (
            "0 * * * * source some-file ; /path/to/some/script.sh",
            {
                "command": "source some-file ; /path/to/some/script.sh",
                "day": "*",
                "hour": "*",
                "minute": "0",
                "month": "*",
                "source": "/etc/crontab",
                "weekday": "*",
            },
            None,
        ),
        (
            r"0 0 * * * sleep ${RANDOM:0:1} && /path/to/executable",
            {
                "command": r"sleep ${RANDOM:0:1} && /path/to/executable",
                "day": "*",
                "hour": "0",
                "minute": "0",
                "month": "*",
                "source": "/etc/crontab",
                "weekday": "*",
            },
            None,
        ),
        (
            "*/5 * * * * /bin/bash -c 'source /some-file; echo \"FOO: $BAR\" >> /var/log/some.log 2>&1'",
            {
                "command": "/bin/bash -c 'source /some-file; echo \"FOO: $BAR\" >> /var/log/some.log 2>&1'",
                "day": "*",
                "hour": "*",
                "minute": "*/5",
                "month": "*",
                "source": "/etc/crontab",
                "weekday": "*",
            },
            None,
        ),
        (
            "0 0 * * * example.sh",
            {
                "command": "example.sh",
                "day": "*",
                "hour": "0",
                "minute": "0",
                "month": "*",
                "source": "/etc/crontab",
                "weekday": "*",
            },
            None,
        ),
        (
            "0 0 * * * root example.sh",
            {
                "command": "root example.sh",
                "day": "*",
                "hour": "0",
                "minute": "0",
                "month": "*",
                "source": "/etc/crontab",
                "weekday": "*",
            },
            "root",
        ),
        (
            "0 0 * * * root\texample.sh",
            {
                "command": "example.sh",
                "day": "*",
                "hour": "0",
                "minute": "0",
                "month": "*",
                "source": "/etc/crontab",
                "weekday": "*",
            },
            "root",
        ),
    ],
)
def test_unix_cronjobs_fuzz(
    cron_line: str,
    expected_output: dict,
    expected_user: str | None,
    target_unix_users: Target,
    fs_unix: VirtualFilesystem,
) -> None:
    """Test if we can handle different cronjob line formats without breaking."""

    fs_unix.map_file_fh("/etc/crontab", BytesIO(cron_line.encode()))
    results = list(target_unix_users.cronjobs())
    assert len(results) == 1
    assert {
        k: v for k, v in results[0]._asdict().items() if k in [f for _, f in CronjobRecord.target_fields]
    } == expected_output
    assert results[0].username == expected_user



from __future__ import annotations

from pathlib import Path

from dissect.target.helpers import keychain
from tests._utils import absolute_path


def test_keychain_register_keychain_file() -> None:
    keychain_file = Path(absolute_path("_data/helpers/keychain/keychain.csv"))

    keychain.register_keychain_file(keychain_file)

    assert len(keychain.get_keys_without_provider()) == 1
    assert len(keychain.get_keys_for_provider("some")) == 0
    assert len(keychain.get_keys_for_provider("bitlocker")) == 2


def test_keychain_register_wildcard_value() -> None:
    keychain.register_wildcard_value("test-value")

    # Number of keys registered is equal number of supported key types, minus one for an invalid raw key
    assert len(keychain.get_keys_without_provider()) == len(keychain.KeyType) - 1

    keychain.KEYCHAIN.clear()
    keychain.register_wildcard_value("0000")

    # Valid raw key now included
    assert len(keychain.get_keys_without_provider()) == len(keychain.KeyType)

    keychain.KEYCHAIN.clear()
    keychain.register_wildcard_value("'312268-409816-583517-486695-627121-599511-664389-145640'")
    for key in keychain.get_all_keys():
        if key.key_type in (keychain.KeyType.RECOVERY_KEY, keychain.KeyType.FILE):
            assert key.value == "312268-409816-583517-486695-627121-599511-664389-145640"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct

from dissect.target.exceptions import RegistryValueNotFoundError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, alias, export

if TYPE_CHECKING:
    from collections.abc import Iterator

pid_def = """
struct DigitalProductId {
    DWORD   uiSize;
    SHORT   MajorVersion;
    SHORT   MinorVersion;
    CHAR    szProductId[24];
    DWORD   uiKeyIdx;
    CHAR    szEditionId[16];
    CHAR    bCDKey[16];
    DWORD   uiCloneStatus;
    DWORD   uiTime;
    DWORD   uiRandom;
    DWORD   uiLt;
    DWORD   uiLicenseData[2];
    CHAR    sOemId[8];
    DWORD   uiBundleId;
    CHAR    sHardwareIdStatic[8];
    DWORD   uiHardwareIdTypeStatic;
    DWORD   uiBiosChecksumStatic;
    DWORD   uiVolSerStatic;
    DWORD   uiTotalRamStatic;
    DWORD   uiVideoBiosChecksumStatic;
    CHAR    sHardwareIdDynamic[8];
    DWORD   uiHardwareIdTypeDynamic;
    DWORD   uiBiosChecksumDynamic;
    DWORD   uiVolSerDynamic;
    DWORD   uiTotalRamDynamic;
    DWORD   uiVideoBiosChecksumDynamic;
    DWORD   uiCRC32;
};

struct DigitalProductId4 {
    DWORD   uiSize;
    SHORT   MajorVersion;
    SHORT   MinorVersion;
    WCHAR   szAdvancedPid[64];
    WCHAR   szActivationId[64];
    WCHAR   szOemID[8];
    WCHAR   szEditionType[260];
    BYTE    bIsUpgrade;
    CHAR    bReserved[7];
    CHAR    bCDKey[16];
    CHAR    bCDKey256Hash[32];
    CHAR    b256Hash[32];
    WCHAR   szEditionId[64];
    WCHAR   szKeyType[64];
    WCHAR   szEULA[64];
};
"""

c_pid = cstruct().load(pid_def)

WindowsProductKeyRecord = TargetRecordDescriptor(
    "windows/product_key",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("string", "type"),
        ("string", "key"),
        ("path", "source"),
    ],
)


class WindowsProductKeyPlugin(Plugin):
    """Windows product key plugin."""

    KEY = "HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"

    # Try the more detailed DigitalProductId version 4 first
    VALUES = ("DigitalProductId4", "DigitalProductId")

    def check_compatible(self) -> None:
        if not self.target.has_function("registry") or not any(self.target.registry.values(self.KEY, self.VALUES)):
            raise UnsupportedPluginError("WindowsProductKeyPlugin plugin requires RegistryPlugin")

    @alias("license")
    @export(record=WindowsProductKeyRecord)
    def productkey(self) -> Iterator[WindowsProductKeyRecord]:
        """Yield Windows product key(s) of the target.

        References:
            - Reversing ``pidgen.dll`` and ````pidgenx.dll``.
            - https://www.licenturion.com/xp/fully-licensed-wpa.txt
            - https://github.com/Endermanch/XPKeygen
        """
        seen_keys = set()

        for key in self.target.registry.keys(self.KEY):
            for value in self.VALUES:
                try:
                    pid_value = key.value(value)
                except RegistryValueNotFoundError:
                    continue

                try:
                    s = getattr(c_pid, pid_value.name)(pid_value.value)
                except EOFError as e:
                    self.target.log.warning("Unable to decode %s structure: %s", pid_value.name, e)
                    self.target.log.debug("", exc_info=e)
                    continue

                if (cd_key := decode_cd_key(s.bCDKey)) in seen_keys:
                    continue

                seen_keys.add(cd_key)

                edition_type = getattr(s, "szEditionType", "").strip("\00")
                edition_id = (
                    s.szEditionId.decode().strip("\00")
                    if isinstance(s.szEditionId, bytes)
                    else s.szEditionId.strip("\00")
                )

                yield WindowsProductKeyRecord(
                    ts=key.ts,
                    name=f"{edition_type} {edition_id}".strip(),
                    type=getattr(s, "szKeyType", "").strip("\00"),
                    key=cd_key,
                    source=f"HKLM\\{key.path}\\{value}",
                    _target=self.target,
                )


def decode_cd_key(input: bytes | bytearray) -> str:
    """Decode the given bytes to a Windows CD key using a pseudo base24 implementation."""
    chars = "BCDFGHJKMPQRTVWXY2346789"
    output = ""

    if not isinstance(input, bytearray):
        input = bytearray(input)

    if is_win8 := (input[14] // 6) & 1:
        input[14] = (input[14] & 0xF7) | ((is_win8 & 2) * 4)

    for _ in range(25):
        c = 0
        for j in range(14, -1, -1):
            c = (c << 8) ^ input[j]
            input[j] = min(int(c / 24), 255)
            c = c % 24
        output = chars[c] + output

    if is_win8:
        if c == 0:
            output = "N" + output
        else:
            part = output[1 : 1 + c]
            output = output[1:].replace(part, part + "N")

    return "-".join([output[i * 5 : i * 5 + 5] for i in range(5)])



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, Callable

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loader import open as loader_open
from dissect.target.loaders.overlay import OverlayLoader
from dissect.target.target import Target

if TYPE_CHECKING:
    from pathlib import Path

BASE_PATH = "/home/user/.local/share/containers/storage/overlay/f351129587e2bb1da9ba4f03dcd22e1c838cd4f20dcc70e6da72381d2905b913"  # noqa: E501


@pytest.fixture
def mock_oci_podman_fs() -> VirtualFilesystem:
    vfs = VirtualFilesystem()

    vfs.makedirs(BASE_PATH)
    vfs.makedirs(f"{BASE_PATH}/diff")
    vfs.makedirs(f"{BASE_PATH}/work")
    vfs.map_file_fh(f"{BASE_PATH}/link", BytesIO())
    vfs.map_file_fh(f"{BASE_PATH}/lower", BytesIO())

    vfs.map_file_fh(f"{BASE_PATH}/diff/test.txt", BytesIO(b"example"))

    return vfs


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], mock_oci_podman_fs: VirtualFilesystem) -> None:
    """Test that we correctly use ``OverlayLoader`` when opening a ``Target``."""
    path = mock_oci_podman_fs.path(BASE_PATH)
    target = opener(path)
    assert isinstance(target._loader, OverlayLoader)
    assert target.path == path


def test_oci_podman(mock_oci_podman_fs: VirtualFilesystem) -> None:
    """Test if we correctly detect and map a Podman OCI container."""
    loader = loader_open(mock_oci_podman_fs.path(BASE_PATH))
    assert isinstance(loader, OverlayLoader)

    t = Target()
    loader.map(t)
    t.apply()

    assert len(t.filesystems) == 1
    assert list(map(str, t.fs.path("/").iterdir())) == ["/test.txt"]



from __future__ import annotations

import logging
import stat
import tarfile as tf
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import BufferedStream

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import (
    Filesystem,
    FilesystemEntry,
    VirtualDirectory,
    VirtualFile,
    VirtualFilesystem,
)
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

log = logging.getLogger(__name__)


class TarFilesystem(Filesystem):
    """Filesystem implementation for tar files."""

    __type__ = "tar"

    def __init__(
        self,
        fh: BinaryIO,
        base: str | None = None,
        *,
        tarinfo: tf.TarInfo | None = None,
        tarfile: tf.TarFile | None = None,
        **kwargs,
    ):
        super().__init__(fh, **kwargs)

        if tarfile:
            self.tar = tarfile
        else:
            fh.seek(0)
            self.tar = tf.open(mode="r", fileobj=fh, tarinfo=tarinfo)  # noqa: SIM115

        self.base = base or ""

        self._fs = VirtualFilesystem(alt_separator=self.alt_separator, case_sensitive=self.case_sensitive)

        for member in self.tar.getmembers():
            mname = member.name.strip("/")
            if not mname.startswith(self.base) or mname == ".":
                continue

            rel_name = fsutil.normpath(mname[len(self.base) :], alt_separator=self.alt_separator)

            entry_cls = TarFilesystemDirectoryEntry if member.isdir() else TarFilesystemEntry
            file_entry = entry_cls(self, rel_name, member)
            self._fs.map_file_entry(rel_name, file_entry)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detect a tar file on a given file-like object."""
        fh = fsutil.open_decompress(fileobj=fh)

        fh.seek(257)
        return fh.read(8) in (tf.GNU_MAGIC, tf.POSIX_MAGIC)

    def get(self, path: str, relentry: FilesystemEntry | None = None) -> FilesystemEntry:
        """Returns a TarFilesystemEntry object corresponding to the given path."""
        return self._fs.get(path, relentry=relentry)


class TarFilesystemEntry(VirtualFile):
    def open(self) -> BinaryIO:
        """Returns file handle (file-like object)."""
        if self.is_dir():
            raise IsADirectoryError(self.path)

        try:
            f = self.fs.tar.extractfile(self.entry)
            if hasattr(f, "raw"):
                f.size = f.raw.size
            return BufferedStream(f, size=f.size)
        except Exception:
            raise FileNotFoundError

    def iterdir(self) -> Iterator[str]:
        if self.is_dir():
            return self._resolve().iterdir()
        return super().iterdir()

    def scandir(self) -> Iterator[FilesystemEntry]:
        if self.is_dir():
            return self._resolve().scandir()
        return super().scandir()

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a directory."""
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.isdir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a file."""
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.isfile()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        """Return whether this entry is a link."""
        return self.entry.issym()

    def readlink(self) -> str:
        """Read the link if this entry is a symlink. Returns a string."""
        if not self.is_symlink():
            raise NotASymlinkError
        return self.entry.linkname

    def readlink_ext(self) -> FilesystemEntry:
        """Read the link if this entry is a symlink. Returns a filesystem entry."""
        # Can't use the one in VirtualFile as it overrides the FilesystemEntry
        return fsutil.resolve_link(self.fs, self.readlink(), self.path, alt_separator=self.fs.alt_separator)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        return fsutil.stat_result(
            [
                (stat.S_IFLNK if self.entry.issym() else stat.S_IFREG) | self.entry.mode,
                self.entry.offset,
                id(self.fs),
                0,
                self.entry.uid,
                self.entry.gid,
                self.entry.size,
                0,
                self.entry.mtime,
                0,
            ]
        )


class TarFilesystemDirectoryEntry(VirtualDirectory):
    def __init__(self, fs: TarFilesystem, path: str, entry: tf.TarInfo):
        super().__init__(fs, path)
        self.entry = entry

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        return fsutil.stat_result(
            [
                stat.S_IFDIR | self.entry.mode,
                self.entry.offset,
                id(self.fs),
                1,
                self.entry.uid,
                self.entry.gid,
                self.entry.size,
                0,
                self.entry.mtime,
                0,
            ]
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.child.vmware_workstation import (
    VmwareWorkstationChildTargetPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("target", "fs", "inventory_path"),
    [
        ("target_win_users", "fs_win", "Users\\John\\AppData\\Roaming\\VMware\\inventory.vmls"),
        ("target_unix_users", "fs_unix", "/home/user/.vmware/inventory.vmls"),
    ],
)
def test_child_vmware_workstation(
    target: Target, fs: VirtualFilesystem, inventory_path: str, request: pytest.FixtureRequest
) -> None:
    """Test if we detect VMware Workstation children from inventory files correctly on Windows and Unix targets."""

    target = request.getfixturevalue(target)
    fs = request.getfixturevalue(fs)

    fs.map_file(inventory_path, absolute_path("_data/plugins/child/vmware_workstation/inventory.vmls"))
    target.add_plugin(VmwareWorkstationChildTargetPlugin)
    children = list(target.list_children())

    assert len(children) == 3
    assert [c.path for c in children] == [
        "/path/to/first/vm/vm.vmx",
        "/path/to/second/vm/vm.vmx",
        "/path/to/third/vm/vm.vmx",
    ]



from __future__ import annotations

import pytest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.regutil import RegistryHive, VirtualKey
from dissect.target.plugins.os.windows.regf.appxdebugkeys import AppxDebugKeysPlugin
from dissect.target.target import Target


def strip_hive(key_path: str) -> str:
    _, _, path = key_path.partition("\\")
    return path


PACKAGED_APPX_KEY_NAME = strip_hive(
    AppxDebugKeysPlugin.REGKEY_GLOBS[0].replace(
        "*",
        "Some.AppX.Package_1.10.3",
        1,
    )
)
PACKAGED_APPX_VALUE_NAME = "(Default)"

ACTIVATABLE_CLASSES_KEY_NAME = strip_hive(
    AppxDebugKeysPlugin.REGKEY_GLOBS[1]
    .replace(
        "*",
        "Some.AppX.Package_1.10.3",
        1,
    )
    .replace(
        "*",
        "Some.AppX.Package.Component.AppX",
        1,
    ),
)
ACTIVATABLE_CLASSES_VALUE_NAME = "DebugPath"

RANDOM_KEY_NAME = f"{ACTIVATABLE_CLASSES_KEY_NAME}\\Random"

DEBUG_INFO = "C:\\windows\\system32\\cmd.exe"


@pytest.fixture
def target_win_appx(target_win_users: Target, hive_hku: RegistryHive) -> Target:
    packaged_appx_key = VirtualKey(hive_hku, PACKAGED_APPX_KEY_NAME)
    packaged_appx_key.add_value(PACKAGED_APPX_VALUE_NAME, DEBUG_INFO)
    hive_hku.map_key(PACKAGED_APPX_KEY_NAME, packaged_appx_key)

    activatable_classes_key = VirtualKey(hive_hku, ACTIVATABLE_CLASSES_KEY_NAME)
    activatable_classes_key.add_value(ACTIVATABLE_CLASSES_VALUE_NAME, DEBUG_INFO)
    hive_hku.map_key(ACTIVATABLE_CLASSES_KEY_NAME, activatable_classes_key)

    random_key = VirtualKey(hive_hku, RANDOM_KEY_NAME)
    hive_hku.map_key(RANDOM_KEY_NAME, random_key)

    return target_win_users


def test_appx_debug_keys_plugin_walk(target_win_appx: Target, hive_hku: RegistryHive) -> Target:
    appx_plugin = AppxDebugKeysPlugin(target_win_appx)
    key = hive_hku.key(ACTIVATABLE_CLASSES_KEY_NAME)
    records = list(appx_plugin._walk(key))

    assert len(records) == 2

    assert records[0].name == ACTIVATABLE_CLASSES_VALUE_NAME
    assert records[0].debug_info == DEBUG_INFO

    # Empty keys should also be returned
    assert records[1].name is None
    assert records[1].debug_info is None


def test_appx_debug_keys__debug_keys(target_win_appx: Target) -> None:
    appx_plugin = AppxDebugKeysPlugin(target_win_appx)
    records = list(appx_plugin._debug_keys())

    assert len(records) == 3


def test_appx_debug_keys_check_compatible(target_win_appx: Target) -> None:
    appx_plugin = AppxDebugKeysPlugin(target_win_appx)
    appx_plugin.check_compatible()


def test_appx_debug_keys_check_compatible_fail() -> None:
    appx_plugin = AppxDebugKeysPlugin(Target())
    with pytest.raises(UnsupportedPluginError):
        appx_plugin.check_compatible()


def test_appx_debug_keys_appxdebugkeys(target_win_appx: Target) -> None:
    appx_plugin = AppxDebugKeysPlugin(target_win_appx)
    records = list(appx_plugin.appxdebugkeys())

    assert len(records) == 3

    assert records[0].name == PACKAGED_APPX_VALUE_NAME
    assert records[0].debug_info == DEBUG_INFO

    assert records[1].name == ACTIVATABLE_CLASSES_VALUE_NAME
    assert records[1].debug_info == DEBUG_INFO

    # Empty keys should also be returned
    assert records[2].name is None
    assert records[2].debug_info is None



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING, Callable
from unittest.mock import MagicMock, patch

import pytest

from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.fixture
def mock_cbc_sdk(monkeypatch: pytest.MonkeyPatch) -> Iterator[MagicMock]:
    with monkeypatch.context() as m:
        # The references to cbc_sdk properties in the cb loader will point to the MagicMock created
        # for the first test function that runs.
        # Thus we need to delete the cb loader module to force it to be reimported when used in a
        # new test so the MagickMock used in the cb module will be the one created for the test
        # function that is running.
        if "dissect.target.loaders.cb" in sys.modules:
            m.delitem(sys.modules, "dissect.target.loaders.cb")
        if "dissect.target.filesystems.cb" in sys.modules:
            m.delitem(sys.modules, "dissect.target.filesystems.cb")

        mock_cbc_sdk = MagicMock()
        m.setitem(sys.modules, "cbc_sdk", mock_cbc_sdk)
        m.setitem(sys.modules, "cbc_sdk.errors", mock_cbc_sdk.errors)
        m.setitem(sys.modules, "cbc_sdk.live_response_api", mock_cbc_sdk.live_response_api)
        m.setitem(sys.modules, "cbc_sdk.platform", mock_cbc_sdk.platform)
        m.setitem(sys.modules, "cbc_sdk.rest_api", mock_cbc_sdk.rest_api)

        yield mock_cbc_sdk


@pytest.fixture
def mock_device(mock_cbc_sdk: MagicMock) -> MagicMock:
    mock_device = MagicMock()
    mock_device.name = "DOMAIN\\workstation"

    mock_cbc_sdk.rest_api.CBCloudAPI.return_value.select.return_value.all.return_value = [mock_device]

    return mock_device


@pytest.fixture
def mock_session(mock_device: MagicMock) -> MagicMock:
    mock_session = MagicMock()
    mock_device.lr_session.return_value = mock_session

    return mock_session


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], mock_session: MagicMock) -> None:
    """Test that we correctly use ``CbLoader`` when opening a ``Target``."""
    from dissect.target.loaders.cb import CbLoader

    path = "cb://instance/workstation"

    with patch("dissect.target.target.Target.apply"):
        target = opener(path)
        assert isinstance(target._loader, CbLoader)
        assert target.path == Path("instance/workstation")


def test_loader(mock_session: MagicMock, mock_cbc_sdk: MagicMock) -> None:
    """Test the CB loader."""
    from dissect.target.filesystems.cb import CbFilesystem
    from dissect.target.loader import open as loader_open
    from dissect.target.loaders.cb import CbLoader, CbRegistry

    loader = loader_open("cb://instance/workstation")
    assert isinstance(loader, CbLoader)
    assert loader.session is mock_session
    assert loader.host == "workstation"
    mock_cbc_sdk.rest_api.CBCloudAPI.assert_called_with(profile="instance")

    mock_session.session_data = {"drives": ["C:\\"]}

    t = Target()
    loader.map(t)

    assert len(t.filesystems) == 1
    assert isinstance(t.fs.mounts["c:\\"], CbFilesystem)
    assert t.fs.mounts["c:\\"].prefix == "c:\\"

    assert len(t._plugins) == 1
    assert isinstance(t._plugins[0], CbRegistry)


def test_registry(target_bare: Target, mock_session: MagicMock) -> None:
    """Test the CB registry plugin."""
    from dissect.target.loaders.cb import CbRegistry

    mock_session.list_registry_keys_and_values.return_value = {"sub_keys": ["TestKey"]}

    registry = CbRegistry(target_bare, mock_session)

    test_key = registry.key("HKLM\\SYSTEM\\TestKey")
    assert test_key.name == "TestKey"
    assert test_key.path == "HKEY_LOCAL_MACHINE\\SYSTEM\\TestKey"
    assert str(test_key.timestamp) == "1970-01-01 00:00:00+00:00"

    mock_session.list_registry_keys_and_values.return_value = {
        "sub_keys": ["TestSubKey"],
        "values": [
            {
                "registry_name": "TestValue",
                "registry_data": None,
                "registry_type": "pbREG_NONE",
            },
            {
                "registry_name": "TestBinary",
                "registry_data": "7a6f6d67",
                "registry_type": "pbREG_BINARY",
            },
            {
                "registry_name": "TestDword",
                "registry_data": "1768515945",
                "registry_type": "pbREG_DWORD",
            },
            {
                "registry_name": "TestQword",
                "registry_data": "7595718147998050665",
                "registry_type": "pbREG_QWORD",
            },
            {
                "registry_name": "TestMultiString",
                "registry_data": "zomg,bbq,wowzers",
                "registry_type": "pbREG_MULTI_SZ",
            },
            {
                "registry_name": "TestString",
                "registry_data": "boring",
                "registry_type": "pbREG_SZ",
            },
        ],
    }

    assert len(test_key.subkeys()) == 1
    assert test_key.subkey("TestSubKey")
    mock_session.list_registry_keys_and_values.assert_called_with("HKEY_LOCAL_MACHINE\\SYSTEM\\TestKey")

    assert len(test_key.values()) == 6
    test_value = test_key.value("TestValue")
    assert test_value.name == "TestValue"
    assert test_value.value is None
    assert test_value.type == "pbREG_NONE"

    assert test_key.value("TestBinary").value == b"zomg"
    assert test_key.value("TestDword").value == 0x69696969
    assert test_key.value("TestQword").value == 0x6969696969696969
    assert test_key.value("TestMultiString").value == ["zomg", "bbq", "wowzers"]
    assert test_key.value("TestString").value == "boring"



from __future__ import annotations

import re
from itertools import chain
from typing import TYPE_CHECKING

from dissect.target.exceptions import LoaderError
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.filesystems.vbk import VbkFilesystem
from dissect.target.loader import Loader, find_loader
from dissect.target.loaders.raw import RawLoader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


RE_RAW_DISK = re.compile(r"(?:[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12})|(?:DEV__.+)")


class VbkLoader(Loader):
    """Load Veaam Backup (VBK) files.

    References:
        - https://helpcenter.veeam.com/docs/backup/hyperv/backup_files.html?ver=120
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self.vbkfs = VbkFilesystem(path.open("rb"))
        self.loader = None

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".vbk"

    def map(self, target: Target) -> None:
        # We haven't really researched any of the VBK metadata yet, so just try some common formats
        root = self.vbkfs.path("/")
        if (base := next(root.glob("*"), None)) is None:
            raise LoaderError("Unexpected empty VBK filesystem")

        if not (
            candidates := [path for pattern in ("*.vmx", "Config/*.vmcx") if (path := next(base.glob(pattern), None))]
        ):
            # Try to look for raw disks
            if not (disks := [path for path in base.iterdir() if RE_RAW_DISK.match(path.name)]):
                # Dunno, just give up 🤷‍♂️ should've spent extra time staring at summary.xml
                raise LoaderError("Unsupported VBK structure, use `-L raw` to manually inspect the VBK")

            candidates.append(root.joinpath("+".join(map(str, disks))))

        # Try to find a loader
        for candidate in candidates:
            if candidate.suffix.lower() == ".vmcx":
                # For VMCX files we need to massage the file layout a bit
                vfs = VirtualFilesystem()
                vfs.map_file_entry(candidate.name, candidate)

                for entry in chain(base.glob("Ide*/*"), base.glob("Scsi*/*")):
                    vfs.map_file_entry(entry.name, entry)

                candidate = vfs.path(candidate.name)

            if (loader := find_loader(candidate, fallbacks=[RawLoader])) is not None:
                ldr = loader(candidate)
                ldr.map(target)

                # Store a reference to the loader if we successfully mapped
                self.loader = ldr
                break
        else:
            raise LoaderError("Unsupported VBK structure, use `-L raw` to manually inspect the VBK")



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.cellebrite import CellebriteFilesystem, CellebriteLoader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``CellebriteLoader`` when opening a ``Target``."""
    path = absolute_path("_data/loaders/cellebrite/EvidenceCollection.ufdx")

    target = opener(path)
    assert isinstance(target._loader, CellebriteLoader)
    assert target.path == path


def test_loader() -> None:
    """Test if we correctly detect and load a Cellebrite UFDX FFS zip extraction from DigitalCorpora (``iOS_17``).

    The content of the ``EXTRACTION_FFS.zip`` file has been replaced with a minimal linux folder structure
    for performance reasons.

    References:
        - https://corp.digitalcorpora.org/corpora/iOS17/
    """
    path = absolute_path("_data/loaders/cellebrite/EvidenceCollection.ufdx")

    loader = loader_open(path)
    assert isinstance(loader, CellebriteLoader)

    assert loader.ufdx.path.name == "EvidenceCollection.ufdx"
    assert loader.ufdx.evidence == "00f905b7-5131-42d9-9ccf-2227115d9536"
    assert loader.ufdx.device.vendor == "Apple"
    assert loader.ufdx.device.model == "iPhone 11 (N104AP)"
    assert loader.ufdx.device.fguid == "98ec76e1-a885-4733-91dd-5dbb61156335"
    assert loader.ufdx.device.guid == "98ec76e1-a885-4733-91dd-5dbb61156335"
    assert not loader.ufdx.device.os
    assert len(loader.ufdx.extractions) == 1
    assert loader.ufdx.extractions[0].type == "FileSystemDump"
    assert loader.ufdx.extractions[0].path.name == "EXTRACTION_FFS.ufd"

    assert len(loader.ufd) == 1
    assert loader.ufd[0].path.name == "EXTRACTION_FFS.ufd"
    assert loader.ufd[0].device.vendor == "Apple"
    assert loader.ufd[0].device.model == "iPhone 11 (N104AP)"
    assert not loader.ufd[0].device.fguid
    assert not loader.ufd[0].device.guid
    assert loader.ufd[0].device.os == "17.3 (21D50)"
    assert len(loader.ufd[0].dumps) == 2
    assert loader.ufd[0].dumps[0].type == "FileDump"
    assert loader.ufd[0].dumps[0].path.name == "EXTRACTION_FFS.zip"
    assert loader.ufd[0].dumps[1].type == "Keychain"

    t = Target()
    loader.map(t)
    t.apply()

    assert len(t.filesystems) == 1
    assert isinstance(t.filesystems[0], CellebriteFilesystem)

    assert sorted(map(str, t.fs.path("/").iterdir())) == [
        "/$fs$",
        "/etc",
        "/home",
        "/opt",
        "/root",
        "/var",
    ]

    assert t.os == "linux"
    assert t.hostname == "example"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystem import Filesystem
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.bsd._os import BsdPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class OpenBsdPlugin(BsdPlugin):
    def __init__(self, target: Target):
        super().__init__(target)
        self._hostname, self._domain = self._parse_hostname_string([("/etc/myname", None)])

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        OPENBSD_PATHS = {
            "/bsd",
            "/bsd.rd",
            "/bsd.mp",
        }

        for fs in target.filesystems:
            if any(fs.exists(path) for path in OPENBSD_PATHS):
                return fs
        return None

    @export(property=True)
    def version(self) -> str | None:
        return None

    @export(property=True)
    def hostname(self) -> str | None:
        return self._hostname

    @export(property=True)
    def domain(self) -> str | None:
        return self._domain



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.plugins.os.unix.linux.proc import (
        NetSocket,
        PacketSocket,
        UnixSocket,
    )
    from dissect.target.target import Target

NetSocketRecord = TargetRecordDescriptor(
    "linux/proc/socket/net",
    [
        ("string", "protocol"),
        ("uint32", "rx_queue"),
        ("uint32", "tx_queue"),
        ("net.ipaddress", "local_ip"),
        ("uint16", "local_port"),
        ("net.ipaddress", "remote_ip"),
        ("uint16", "remote_port"),
        ("string", "state"),
        ("string", "owner"),
        ("uint32", "inode"),
        ("uint32", "pid"),
        ("string", "name"),
        ("string", "cmdline"),
    ],
)

UnixSocketRecord = TargetRecordDescriptor(
    "linux/proc/socket/unix",
    [
        ("string", "protocol"),
        ("uint32", "ref"),
        ("string", "socket_flags"),
        ("string", "type"),
        ("string", "state"),
        ("uint32", "inode"),
        ("string", "path"),
    ],
)

PacketSocketRecord = TargetRecordDescriptor(
    "linux/proc/socket/packet",
    [
        ("string", "protocol"),
        ("string", "protocol_type"),
        ("uint32", "socket_type"),
        ("uint32", "sk"),
        ("uint32", "ref"),
        ("uint32", "iface"),
        ("uint32", "r"),
        ("uint32", "rmem"),
        ("uint32", "uid"),
        ("string", "owner"),
        ("uint32", "inode"),
        ("uint32", "pid"),
        ("string", "name"),
        ("string", "cmdline"),
    ],
)


class NetSocketPlugin(Plugin):
    """Linux volatile net sockets plugin."""

    __namespace__ = "sockets"

    def __init__(self, target: Target):
        super().__init__(target)
        self.sockets = self.target.proc.sockets

    def check_compatible(self) -> None:
        if not self.target.has_function("proc"):
            raise UnsupportedPluginError("proc filesystem not available")

    @export(record=PacketSocketRecord)
    def packet(self) -> Iterator[PacketSocketRecord]:
        """This plugin yields the packet sockets and available stats associated with them.

        Yields PacketSocketRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            protocol (str): packet.
            protocol_type (str): The canonical name of the captured protocol i.e. ETH_P_ALL.
            socket_type (int): The integer type of the socket (packet).
            sk (int): The socket number.
            iface (int): The interface index of the socket.
            r (int): The number of bytes that have been received by the socket and are waiting to be processed.
            rmem (int): The size of the receive buffer for the socket.
            uid (int): The user ID of the process that created the socket.
            inode (int): The inode associated to the socket.
            pid (int): The pid associated with this socket.
            name (string): The process name associated to this socket.
            cmdline (string): The command line used to start the socket with.
            owner (string): The resolved user ID of the socket.
        """
        yield from map(self._generate_packet_socket_record, self.sockets.packet())

    @export(record=UnixSocketRecord)
    def unix(self) -> Iterator[UnixSocketRecord]:
        """This plugin yields the unix sockets and available stats associated with them.

        Yields UnixSocketRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            protocol (string): The protocol used by the socket.
            socket_flags (bytes): The flags associated with the socket.
            type (string): The stream type of the socket.
            state (string): The state of the socket.
            inode (int): The inode associated to the socket.
            path (string): The path associated to the socket.
        """
        yield from map(self._generate_unix_socket_record, self.sockets.unix())

    @export(record=NetSocketRecord)
    def raw(self) -> Iterator[NetSocketRecord]:
        """This plugin yields the raw and raw6 sockets and available stats associated with them.

        Yields NetSocketRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            protocol (string): The protocol used by the socket.
            receive_queue (int): The size, in bytes of the receive queue of the socket.
            transmit_queue (int): The size, in bytes of the transmit queue of the socket.
            local_ip (string): The local ip the socket connects from.
            local_port (int): The local port the socket connects from.
            remote_ip (string): The remote ip the socket connects to.
            remote_port (int): The remote port the socket connects to.
            state (string): The state of the socket.
            owner (string): The loginuid of the pid associated with this socket.
            inode (int): The inode (fd) associated with this socket.
            pid (int): The pid associated with this socket.
            name (string): The process name associated with this socket.
            cmdline (string): The command line used to start the socket with.
        """
        yield from map(self._generate_net_socket_record, chain(self.sockets.raw(), self.sockets.raw6()))

    @export(record=NetSocketRecord)
    def udp(self) -> Iterator[NetSocketRecord]:
        """This plugin yields the udp and udp6 sockets and available stats associated with them.

        Yields NetSocketRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            protocol (string): The protocol used by the socket.
            receive_queue (int): The size, in bytes of the receive queue of the socket.
            transmit_queue (int): The size, in bytes of the transmit queue of the socket.
            local_ip (string): The local ip the socket connects from.
            local_port (int): The local port the socket connects from.
            remote_ip (string): The remote ip the socket connects to.
            remote_port (int): The remote port the socket connects to.
            state (string): The state of the socket.
            owner (string): The loginuid of the pid associated with this socket.
            inode (int): The inode (fd) associated with this socket.
            pid (int): The pid associated with this socket.
            name (string): The process name associated with this socket.
            cmdline (string): The command line used to start the socket with.
        """
        yield from map(self._generate_net_socket_record, chain(self.sockets.udp(), self.sockets.udp6()))

    @export(record=NetSocketRecord)
    def tcp(self) -> Iterator[NetSocketRecord]:
        """This plugin yields the tcp and tcp6 sockets and available stats associated with them.

        Yields NetSocketRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            protocol (string): The protocol used by the socket.
            receive_queue (int): The size, in bytes of the receive queue of the socket.
            transmit_queue (int): The size, in bytes of the transmit queue of the socket.
            local_ip (string): The local ip the socket connects from.
            local_port (int): The local port the socket connects from.
            remote_ip (string): The remote ip the socket connects to.
            remote_port (int): The remote port the socket connects to.
            state (string): The state of the socket.
            owner (string): The loginuid of the pid associated with this socket.
            inode (int): The inode (fd) associated with this socket.
            pid (int): The pid associated with this socket.
            name (string): The process name associated with this socket.
            cmdline (string): The command line used to start the socket with.
        """
        yield from map(self._generate_net_socket_record, chain(self.sockets.tcp(), self.sockets.tcp6()))

    def _generate_unix_socket_record(self, data: UnixSocket) -> UnixSocketRecord:
        return UnixSocketRecord(
            protocol=data.protocol_string,
            ref=data.ref,
            socket_flags=data.flags,
            type=data.stream_type_string,
            state=data.state_string,
            inode=data.inode,
            path=data.path,
            _target=self.target,
        )

    def _generate_packet_socket_record(self, data: PacketSocket) -> PacketSocketRecord:
        return PacketSocketRecord(
            protocol=data.protocol_string,
            protocol_type=data.protocol_type,
            socket_type=data.type,
            sk=data.sk,
            ref=data.ref,
            iface=data.iface,
            r=data.r,
            rmem=data.rmem,
            uid=data.user,
            inode=data.inode,
            pid=data.pid,
            name=data.name,
            cmdline=data.cmdline,
            owner=data.owner,
            _target=self.target,
        )

    def _generate_net_socket_record(self, data: NetSocket) -> NetSocketRecord:
        return NetSocketRecord(
            protocol=data.protocol_string,
            rx_queue=data.rx_queue,
            tx_queue=data.tx_queue,
            local_ip=data.local_ip,
            local_port=data.local_port,
            remote_ip=data.remote_ip,
            remote_port=data.remote_port,
            state=data.state_string,
            owner=data.owner,
            inode=data.inode,
            pid=data.pid,
            name=data.name,
            cmdline=data.cmdline,
            _target=self.target,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.child.virtuozzo import VirtuozzoChildTargetPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_virtuozzo(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.makedirs("vz/root/a")
    fs_unix.makedirs("vz/root/b")

    target_unix.add_plugin(VirtuozzoChildTargetPlugin)

    children = list(target_unix.list_children())
    assert len(children) == 2
    assert children[0].type == "virtuozzo"
    assert str(children[0].path) == "/vz/root/a"
    assert children[1].type == "virtuozzo"
    assert str(children[1].path) == "/vz/root/b"



from __future__ import annotations

import dataclasses
import datetime
import enum
import functools
import json
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, TextIO

import structlog

from dissect.target.tools.dump.utils import (
    Compression,
    Serialization,
    get_current_utc_time,
    parse_datetime_iso,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.target import Target
    from dissect.target.tools.dump.run import RecordStreamElement

log = structlog.get_logger(__name__)


STATE_FILE_NAME = "target-dump.state.json"

PENDING_UPDATES_LIMIT = 10


@dataclass
class Sink:
    target_path: str
    func: str
    path: Path
    is_dirty: bool = True
    record_count: int = 0
    size_bytes: int = 0

    def __post_init__(self):
        self.func = getattr(self.func, "name", self.func)


@dataclass
class DumpState:
    target_paths: list[str]
    functions: str
    excluded_functions: list[str]
    serialization: str
    compression: str
    start_time: datetime.datetime
    last_update_time: datetime.datetime

    sinks: list[Sink] = dataclasses.field(default_factory=list)

    # Volatile properties
    output_dir: Path | None = None
    pending_updates_count: int | None = 0

    @property
    def record_count(self) -> int:
        return sum(s.record_count for s in self.sinks)

    @property
    def finished_sinks(self) -> list[Sink]:
        return [sink for sink in self.sinks if not sink.is_dirty]

    @property
    def path(self) -> Path:
        return DumpState.get_state_path(self.output_dir)

    @classmethod
    def get_state_path(cls, output_dir: Path) -> Path:
        return output_dir / STATE_FILE_NAME

    def get_full_sink_path(self, sink: Sink) -> Path:
        if not self.output_dir:
            raise ValueError("Output directory is unknown for the state")
        return self.output_dir / sink.path

    def get_sink(self, path: Path) -> Sink | None:
        for sink in self.sinks:
            if sink.path == path:
                return sink
        return None

    def serialize(self) -> str:
        """Serialize state instance into a JSON formatted string."""
        state_dict = dataclasses.asdict(self)
        state_dict.pop("output_dir")
        state_dict.pop("pending_updates_count")
        return json.dumps(
            state_dict,
            default=serialize_obj,
            indent=4,
            sort_keys=True,
        )

    def persist(self, fh: TextIO) -> None:
        """Write serialized state instance into profided ``fh`` byte stream, overwriting it from the beginning."""
        fh.seek(0)
        fh.write(self.serialize())
        fh.flush()
        self.pending_updates_count = 0
        log.debug("State flushed")

    def mark_as_finished(self, target: Target, func: str) -> None:
        """Mark sinks that match provided ``target`` and ``func`` pair as not dirty."""
        matching_sinks = [
            sink for sink in self.sinks if str(sink.target_path) == str(target.path) and sink.func == func
        ]
        for sink in matching_sinks:
            sink.is_dirty = False

    def create_sink(self, sink_path: Path, stream_element: RecordStreamElement) -> Sink:
        """Create a sink instance for provided ``sink_path`` and ``stream_element``
        (from which ``target`` and ``func`` properties are used).
        """
        sink = Sink(
            path=sink_path,
            target_path=str(stream_element.target.path),
            func=stream_element.func,
        )
        self.sinks.append(sink)
        return sink

    def update(self, stream_element: RecordStreamElement, fp_position: int) -> None:
        """Update a sink instance for provided ``stream_element``."""
        sink = self.get_sink(stream_element.sink_path)

        sink.record_count += 1
        sink.size_bytes = fp_position
        self.last_update_time = get_current_utc_time()

        self.pending_updates_count += 1
        log.debug("State updated", records=self.record_count, sinks=len(self.sinks), path=self.path)

    @classmethod
    def from_dict(cls, state_dict: dict) -> Self:
        """Deserialize state instance from provided dictionary."""
        return cls(
            target_paths=state_dict["target_paths"],
            functions=state_dict["functions"],
            excluded_functions=state_dict["excluded_functions"],
            serialization=Serialization(state_dict["serialization"]),
            compression=Compression(state_dict["compression"]),
            start_time=parse_datetime_iso(state_dict["start_time"]),
            last_update_time=parse_datetime_iso(state_dict["last_update_time"]),
            sinks=[
                Sink(
                    target_path=sink["target_path"],
                    path=Path(sink["path"]),
                    func=sink["func"],
                    record_count=sink["record_count"],
                    size_bytes=sink["size_bytes"],
                    is_dirty=sink["is_dirty"],
                )
                for sink in state_dict["sinks"]
            ],
        )

    @classmethod
    def from_path(cls, output_dir: Path) -> Self | None:
        """Deserialize state instance from a file in the provided output directory path."""
        state_path = cls.get_state_path(output_dir)
        if not state_path.exists():
            return None

        with state_path.open(mode="r") as fh:
            try:
                state_dict = json.load(fh)
            except ValueError as e:
                log.warning("Can not load state from path", path=state_path, exc=e)
                return None

        state = cls.from_dict(state_dict)
        state.output_dir = output_dir
        return state

    def get_invalid_sinks(self) -> list[Sink]:
        """Return sinks that have a mismatch between recorded size and a real file size."""
        invalid_sinks = []
        for sink in self.sinks:
            # sink file does not exist
            if not self.get_full_sink_path(sink).exists():
                invalid_sinks.append(sink)
                continue

            # recorded file size for a clean sink is incorrect
            if not sink.is_dirty and sink.size_bytes != sink.path.stat().st_size:
                invalid_sinks.append(sink)
                continue

        return invalid_sinks

    def drop_invalid_sinks(self) -> None:
        """Remove sinks that have a mismatch between recorded size and a real file size from the list of sinks."""
        for invalid_sink in self.get_invalid_sinks():
            self.sinks.remove(invalid_sink)
            log.debug("Ignoring invalid sink", sink=invalid_sink.path)

    def drop_dirty_sinks(self) -> None:
        """Drop sinks that are marked as "dirty" in the current state from the list of sinks."""
        dirty_sinks = [s for s in self.sinks if s.is_dirty]
        for dirty_sink in dirty_sinks:
            self.sinks.remove(dirty_sink)
            log.debug("Ignoring dirty sink", sink=dirty_sink.path)


def create_state(
    *,
    output_dir: Path,
    target_paths: list[str],
    functions: str,
    excluded_functions: list[str],
    serialization: Serialization,
    compression: Compression = None,
) -> DumpState:
    """Create a ``DumpState`` instance with provided properties."""
    current_time = get_current_utc_time()
    return DumpState(
        target_paths=target_paths,
        functions=functions,
        excluded_functions=excluded_functions,
        serialization=serialization,
        compression=compression,
        start_time=current_time,
        last_update_time=current_time,
        output_dir=output_dir,
    )


@contextmanager
def persisted_state(state: DumpState) -> Iterator[Callable]:
    """Return a context manager for persisting ``DumpState`` instance."""

    def save_state(fh: TextIO) -> None:
        state.persist(fh)
        state.pending_updates_count = 0

    def save_state_when_over_limit(fh: TextIO) -> None:
        if state.pending_updates_count >= PENDING_UPDATES_LIMIT:
            save_state(fh)

    state.path.parent.mkdir(parents=True, exist_ok=True)

    with state.path.open(mode="w") as fh:
        try:
            yield functools.partial(save_state_when_over_limit, fh)
        finally:
            save_state(fh)


def load_state(output_dir: Path) -> DumpState | None:
    """Load persisted ``DumpState`` instance from provided ``output_dir`` path and perform sink validation."""
    state = DumpState.from_path(output_dir)

    if not state:
        return None

    # Dropping sinks that are marked as finished (clean) but the file
    # size on the disk is different from the one stored in the state object.
    state.drop_invalid_sinks()

    # Since there are no guarantees about the order of produced records
    # and the records themselves do not have unique IDs, we can not easily verify
    # what records are already stored in the dirty sink.
    # It is easier to just restart all dirty sinks.
    state.drop_dirty_sinks()

    return state


def serialize_obj(obj: Any) -> str:
    """JSON serializer for object types not serializable by ``json`` library."""
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    if isinstance(obj, Path):
        return str(obj)
    if isinstance(obj, enum.Enum):
        return obj.value
    raise TypeError(f"Type {type(obj)} not serializable")



from __future__ import annotations

import datetime
import json
import operator
from io import BytesIO
from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import path

from dissect.target.plugins.apps.container.docker import (
    DockerPlugin,
    convert_timestamp,
    find_installs,
    strip_log,
)
from dissect.target.plugins.os.unix._os import UnixPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def target_linux_docker_logs(target_linux: Target, fs_linux: VirtualFilesystem) -> Target:
    docker_containers = absolute_path("_data/plugins/apps/container/docker/logs")
    fs_linux.map_dir("/var/lib/docker/containers", docker_containers)
    return target_linux


def test_docker_plugin_data_roots(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.makedirs("var/lib/docker")
    fs_unix.makedirs("tmp/foo/bar")
    fs_unix.makedirs("tmp/another/docker")
    fs_unix.map_file_fh("/etc/docker/daemon.json", BytesIO(b'{"data-root": "/tmp/foo/bar"}'))
    fs_unix.map_file_fh("/root/.docker/daemon.json", BytesIO(b'{"data-root": "/tmp/another/docker"}'))

    assert [str(p) for p in find_installs(target_unix_users)] == [
        "/var/lib/docker",
        "/tmp/foo/bar",
        "/tmp/another/docker",
    ]


def test_docker_plugin_data_roots_empty(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.makedirs("/var/lib/docker")
    fs_unix.map_file_fh("/etc/docker/daemon.json", BytesIO(b"{}"))

    assert [str(p) for p in find_installs(target_unix_users)] == [
        "/var/lib/docker",
    ]


def test_docker_plugin_images(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file(
        "/var/lib/docker/image/overlay2/repositories.json",
        absolute_path("_data/plugins/apps/container/docker/repositories.json"),
    )

    hash = "6b7dfa7e8fdbe18ad425dd965a1049d984f31cf0ad57fa6d5377cca355e65f03"
    fs_unix.map_file(
        f"/var/lib/docker/image/overlay2/imagedb/content/sha256/{hash}",
        absolute_path("_data/plugins/apps/container/docker/image_metadata.json"),
    )

    target_unix_users.add_plugin(DockerPlugin)
    results = list(target_unix_users.docker.images())
    assert len(results) == 2
    assert results[0].tag == "ubuntu:latest"
    assert results[0].hash == f"sha256:{hash}"
    assert results[0].image_id == "6b7dfa7e8fdb"
    assert results[0].created == datetime.datetime(2022, 12, 9, 1, 20, 31, 321639, tzinfo=datetime.timezone.utc)


def test_docker_plugin_containers(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    id = "d3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33f"

    fs_unix.map_file(
        f"/var/lib/docker/containers/{id}/config.v2.json",
        absolute_path("_data/plugins/apps/container/docker/container_running.json"),
    )
    target_unix_users.add_plugin(DockerPlugin)
    results = list(target_unix_users.docker.containers())

    assert len(results) == 1

    result = results[0]

    assert result.container_id == id
    assert result.image == "exampleimage:1.33.7"
    assert result.image_id == "d3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33fd3adb33f"
    assert result.created == datetime.datetime(2022, 12, 19, 13, 37, 1, 247519, tzinfo=datetime.timezone.utc)
    assert bool(result.running) is True
    assert result.started == datetime.datetime(2022, 12, 19, 13, 37, 1, 247519, tzinfo=datetime.timezone.utc)
    assert result.finished == datetime.datetime(1, 1, 1, 00, 00, 00, 000000, tzinfo=datetime.timezone.utc)
    assert result.ports == [
        "0.0.0.0:1234->1234/tcp",
        ":::1234->1234/tcp",
        "0.0.0.0:5678->5678/tcp",
        ":::5678->5678/tcp",
    ]
    assert result.names == "example_container_name"
    assert result.volumes == ["/tmp/test:/test"]
    assert result.config_path == path.from_posix(f"/var/lib/docker/containers/{id}/config.v2.json")
    assert result.mount_path == path.from_posix(f"/var/lib/docker/image/overlay2/layerdb/mounts/{id}")


def test_docker_plugin_logs(target_linux_docker_logs: Target) -> None:
    target_linux_docker_logs.add_plugin(DockerPlugin)
    results = list(target_linux_docker_logs.docker.logs())
    results.sort(key=operator.attrgetter("ts"))

    assert len(results) == 288

    # json log driver
    assert results[0].ts == datetime.datetime(2023, 11, 9, 8, 42, 57, 79838, tzinfo=datetime.timezone.utc)
    assert results[0].container == "f988f88e221d97930a665712cf16ab520f7e2c5af395660c145df93aebedf071"
    assert results[0].stream == "stdout"
    assert results[0].message == "/ # "

    # local log driver (protobuf)
    assert results[-1].ts == datetime.datetime(2023, 11, 9, 9, 52, 52, 587579, tzinfo=datetime.timezone.utc)
    assert results[-1].container == "0627aa2d32de2478f4a3e8bb3c655ea7baa1a3463d8cee41263655244fe4717c"
    assert results[-1].stream == "stdout"
    assert results[-1].message == "exit"


def test_docker_plugin_logs_raw(target_linux_docker_logs: Target) -> None:
    target_linux_docker_logs.add_plugin(DockerPlugin)
    results = list(target_linux_docker_logs.docker.logs(raw_messages=True))
    results.sort(key=operator.attrgetter("ts"))

    assert len(results) == 288
    assert results[0].message == "/ # \x1b[6n\r\n"
    assert results[-1].message == "\x1b[?2004l\rexit\r"


def test_docker_plugin_timestamps() -> None:
    # Should not alter already correct timestamps
    assert convert_timestamp("2022-12-19T13:37:00.123456") == "2022-12-19T13:37:00.123456"
    assert convert_timestamp("2022-12-19T13:37:00.123456Z") == "2022-12-19T13:37:00.123456Z"

    # Should convert nanosecond timestamps to microsecond timestamps
    assert convert_timestamp("2022-12-19T13:37:00.123456789Z") == "2022-12-19T13:37:00.123456Z"
    assert convert_timestamp("2022-12-19T13:37:00.12345678Z") == "2022-12-19T13:37:00.123456Z"
    assert convert_timestamp("2022-12-19T13:37:00.123456789+01:00") == "2022-12-19T13:37:00.123456+01:00"


def test_backspace_interpretation() -> None:
    # Should 'execute' backspaces found in input
    input = (
        '~ # \x1b[6necho \'\x08\x1b[J"ths \x08\x1b[J\x08\x1b[Js \x08\x1b[J\x08\x1b[Jis is a secret!" > secret.txt\r\n'
    )
    assert strip_log(input, exc_backspace=True) == '~ # echo "this is a secret!" > secret.txt'


def test_regression_running_container_parsing(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we correctly discover and reconstruct exposed container ports and commands on a running container."""

    id = "deadbeef"
    config = {
        "ID": id,
        "Image": "sha256:blabla",
        "Name": "/foo",
        "Path": "/bin/some-binary",
        "Driver": "overlay2",
        "Args": [
            "--some-argument 1",
            "--another-argument 2",
        ],
        "State": {
            "Running": True,
            "Pid": 1337,
            "StartedAt": "2024-12-31T13:37:00.123456789Z",
        },
        "Config": {
            "Hostname": "foo",
            "ExposedPorts": {
                "1337/tcp": {},
                "1337/udp": {},
            },
            "Image": "docker.io/debian",
            "Env": [
                "FOO=bar",
                "HELLO=world",
            ],
        },
        "NetworkSettings": {
            "Ports": {},
        },
        "MountPoints": {
            "/dest/file.txt": {
                "Source": "/somewhere/on/host/file.txt",
                "Destination": "/dest/file.txt",
            },
        },
    }

    fs_unix.map_file_fh(f"/var/lib/docker/containers/{id}/config.v2.json", BytesIO(json.dumps(config).encode()))
    fs_unix.map_file_fh("/etc/hostname", BytesIO(b"hostname"))

    target_unix.add_plugin(UnixPlugin)
    target_unix.add_plugin(DockerPlugin)

    results = list(target_unix.docker.containers())
    assert len(results) == 1

    assert results[0].hostname == "hostname"
    assert results[0].container_id == id
    assert results[0].image == "docker.io/debian"
    assert results[0].image_id == "blabla"
    assert results[0].command == "/bin/some-binary --some-argument 1 --another-argument 2"
    assert results[0].created is None
    assert results[0].running
    assert results[0].pid == 1337
    assert results[0].started == datetime.datetime(2024, 12, 31, 13, 37, 0, 123456, tzinfo=datetime.timezone.utc)
    assert results[0].finished is None
    assert results[0].ports == ["0.0.0.0:1337->1337/tcp", "0.0.0.0:1337->1337/udp"]
    assert results[0].names == "foo"
    assert results[0].volumes == ["/somewhere/on/host/file.txt:/dest/file.txt"]
    assert results[0].environment == ["FOO=bar", "HELLO=world"]
    assert results[0].mount_path == "/var/lib/docker/image/overlay2/layerdb/mounts/deadbeef"
    assert results[0].config_path == "/var/lib/docker/containers/deadbeef/config.v2.json"



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.filesystem import (
    VirtualDirectory,
    VirtualFile,
    VirtualFilesystem,
    VirtualSymlink,
)
from dissect.target.plugins.os.unix.locale import UnixLocalePlugin, timezone_from_path
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_locale_plugin_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    # Locale locations originate from Ubuntu 20.
    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Europe/Amsterdam"))
    fs_unix.map_file_fh("/etc/default/locale", BytesIO(b"LANG=en_US.UTF-8"))
    fs_unix.map_file("/etc/default/keyboard", absolute_path("_data/plugins/os/unix/locale/keyboard"))
    target_unix_users.add_plugin(UnixLocalePlugin)

    assert target_unix_users.timezone == "Europe/Amsterdam"
    assert target_unix_users.language == ["en_US"]
    keyboard = list(target_unix_users.keyboard())
    assert len(keyboard) == 1
    assert keyboard[0].layout == "us"
    assert keyboard[0].model == "pc105"
    assert keyboard[0].variant == ""
    assert keyboard[0].options == ""
    assert keyboard[0].backspace == "guess"


def test_locale_plugin_unix_quotes(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    # Older Fedora system
    fs_unix.map_file_fh("/etc/default/locale", BytesIO(b'LANG="en_US.UTF-8"'))
    target_unix_users.add_plugin(UnixLocalePlugin)

    assert target_unix_users.language == ["en_US"]


def test_locale_etc_localtime_symlink(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.symlink("/usr/share/zoneinfo/Europe/Amsterdam", "/etc/localtime")
    target_unix_users.add_plugin(UnixLocalePlugin)

    assert isinstance(fs_unix.get("/etc/localtime"), VirtualSymlink)
    assert target_unix_users.timezone == "Europe/Amsterdam"


def test_locale_etc_localtime_hardlink(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/usr/share/zoneinfo/Europe/Amsterdam", BytesIO(b"contents of Europe/Amsterdam"))
    fs_unix.link("/usr/share/zoneinfo/Europe/Amsterdam", "/etc/localtime")
    target_unix_users.add_plugin(UnixLocalePlugin)

    assert isinstance(fs_unix.get("/etc/localtime"), VirtualFile)
    assert fs_unix.path("/etc/localtime").samefile(fs_unix.path("/usr/share/zoneinfo/Europe/Amsterdam"))

    entry = fs_unix.get("/etc/localtime")
    stat = entry.stat()
    stat.st_nlink = 2

    with patch.object(entry, "stat", return_value=stat):
        assert entry.stat().st_nlink == 2
        assert target_unix_users.timezone == "Europe/Amsterdam"


def test_locale_etc_localtime_regular_file(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/localtime", BytesIO(b"contents of Europe/Amsterdam"))
    fs_unix.map_file_fh("/usr/share/zoneinfo/UTC", BytesIO(b"contents of UTC"))
    fs_unix.map_file_fh("/usr/share/zoneinfo/Europe/Amsterdam", BytesIO(b"contents of Europe/Amsterdam"))
    fs_unix.map_file_fh("/usr/share/zoneinfo/America/New_York", BytesIO(b"contents of America/New_York"))
    target_unix_users.add_plugin(UnixLocalePlugin)

    assert isinstance(fs_unix.get("/etc/localtime"), VirtualFile)
    assert isinstance(fs_unix.get("/usr/share/zoneinfo/UTC"), VirtualFile)
    assert isinstance(fs_unix.get("/usr/share/zoneinfo/Europe"), VirtualDirectory)
    assert isinstance(fs_unix.get("/usr/share/zoneinfo/Europe/Amsterdam"), VirtualFile)
    assert isinstance(fs_unix.get("/usr/share/zoneinfo/America/New_York"), VirtualFile)
    assert target_unix_users.timezone == "Europe/Amsterdam"


@pytest.mark.parametrize(
    ("input", "expected_output"),
    [
        ("/usr/share/zoneinfo/Europe/Amsterdam", "Europe/Amsterdam"),
        ("/usr/share/zoneinfo/UTC", "UTC"),
        ("Europe/Amsterdam", "Europe/Amsterdam"),
        ("Etc/UTC", "UTC"),
    ],
)
def test_locale_timezone_string_normalize(input: str, expected_output: str) -> None:
    """Test if we normalize zoneinfo paths correctly."""
    assert timezone_from_path(Path(input)) == expected_output



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.loaders.tar import TarLoader
from dissect.target.plugins.filesystem.walkfs import WalkFSPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pytest_benchmark.fixture import BenchmarkFixture

    from dissect.target.target import Target


def test_walkfs_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_entry("/path/to/some/file", VirtualFile(fs_unix, "file", None))
    fs_unix.map_file_entry("/path/to/some/other/file.ext", VirtualFile(fs_unix, "file.ext", None))
    fs_unix.map_file_entry("/root_file", VirtualFile(fs_unix, "root_file", None))
    fs_unix.map_file_entry("/other_root_file.ext", VirtualFile(fs_unix, "other_root_file.ext", None))
    fs_unix.map_file_entry("/.test/test.txt", VirtualFile(fs_unix, "test.txt", None))
    fs_unix.map_file_entry("/.test/.more.test.txt", VirtualFile(fs_unix, ".more.test.txt", None))

    target_unix.add_plugin(WalkFSPlugin)

    results = list(target_unix.walkfs())
    assert len(results) == 14
    assert sorted([r.path for r in results]) == [
        "/",
        "/.test",
        "/.test/.more.test.txt",
        "/.test/test.txt",
        "/etc",
        "/other_root_file.ext",
        "/path",
        "/path/to",
        "/path/to/some",
        "/path/to/some/file",
        "/path/to/some/other",
        "/path/to/some/other/file.ext",
        "/root_file",
        "/var",
    ]


@pytest.mark.benchmark
def test_benchmark_walkfs(target_bare: Target, benchmark: BenchmarkFixture) -> None:
    """Benchmark walkfs performance on a small tar archive with ~500 files."""

    loader = TarLoader(Path(absolute_path("_data/loaders/containerimage/alpine-docker.tar")))
    loader.map(target_bare)
    target_bare.apply()

    result = benchmark(lambda: next(WalkFSPlugin(target_bare).walkfs()))

    assert result.path == "/"



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.target.exceptions import FileNotFoundError, NotADirectoryError
from dissect.target.filesystems.exfat import ExfatFilesystem, ExfatFilesystemEntry

if TYPE_CHECKING:
    from collections.abc import Iterator


def mock_file(name: str, is_dir: bool = False) -> Mock:
    file = Mock(name=name)

    file.metadata = Mock()
    file.metadata.attributes = Mock()
    file.metadata.attributes.directory = is_dir

    file.stream = Mock()
    file.stream.data_length = 0

    return file


MOCK_EXFAT_FILE_TREE = {
    "/": (
        mock_file("Mock_", is_dir=True),
        {
            "some_path": (
                mock_file("Mock_some_path", is_dir=True),
                {
                    "some_file": (
                        mock_file("Mock_some_path_some_file"),
                        None,
                    ),
                    "other_file": (
                        mock_file("Mock_some_path_other_file"),
                        None,
                    ),
                },
            ),
            "other_file": (
                mock_file("Mock_other_file"),
                None,
            ),
        },
    ),
}


@pytest.fixture
def exfat_fs() -> Iterator[ExfatFilesystem]:
    with patch("dissect.fat.exfat.ExFAT"):
        exfat_fs = ExfatFilesystem(Mock())
        exfat_fs.exfat.files = MOCK_EXFAT_FILE_TREE

        yield exfat_fs


@pytest.fixture
def other_file(exfat_fs: ExfatFilesystem) -> ExfatFilesystemEntry:
    return ExfatFilesystemEntry(exfat_fs, "/other_file", MOCK_EXFAT_FILE_TREE["/"][1]["other_file"])


@pytest.fixture
def some_path(exfat_fs: ExfatFilesystem) -> ExfatFilesystemEntry:
    return ExfatFilesystemEntry(exfat_fs, "/some_path", MOCK_EXFAT_FILE_TREE["/"][1]["some_path"])


def test_filesystem_get(exfat_fs: ExfatFilesystem) -> None:
    path = "/some_path/SOME_FILE"
    some_file = exfat_fs.get(path)

    assert isinstance(some_file, ExfatFilesystemEntry)
    assert some_file.path == path
    assert some_file.entry[0]._mock_name == "Mock_some_path_some_file"


def test_filesystem__get_entry(exfat_fs: ExfatFilesystem) -> None:
    root = exfat_fs._get_entry("/")
    assert root[0]._mock_name == "Mock_"
    assert isinstance(root[1], dict)

    some_path = exfat_fs._get_entry("/some_path")

    assert some_path[0]._mock_name == "Mock_some_path"
    assert exfat_fs._get_entry("/SoMe_PaTh") == some_path

    other_file = exfat_fs._get_entry("/some_path/OTHER_FILE")

    assert other_file[0]._mock_name == "Mock_some_path_other_file"
    assert other_file[1] is None
    assert exfat_fs._get_entry("OTHER_FILE", root=some_path) == other_file


def test_filesystem__get_entry_not_a_directory_error(exfat_fs: ExfatFilesystem) -> None:
    with pytest.raises(NotADirectoryError):
        exfat_fs._get_entry("/other_file/non-exisiting_file")


def test_filesystem__get_entry_file_not_found_error(exfat_fs: ExfatFilesystem) -> None:
    with pytest.raises(FileNotFoundError):
        exfat_fs._get_entry("/non-exisiting_file")


def test_filesystem_entry_get(exfat_fs: ExfatFilesystem, some_path: ExfatFilesystemEntry) -> None:
    some_file = some_path.get("some_file")

    assert some_file.path == "/some_path/some_file"
    assert some_file.entry[0]._mock_name == "Mock_some_path_some_file"


def test_filesystem_entry__iterdir(exfat_fs: ExfatFilesystem, some_path: ExfatFilesystemEntry) -> None:
    file_names = []
    file_entries = []

    for entry_name, entry_file_tree in some_path._iterdir():
        file_names.append(entry_name)
        file_entries.append(entry_file_tree[0]._mock_name)

    assert len(file_names) == 2
    assert "some_file" in file_names
    assert "other_file" in file_names
    assert "Mock_some_path_some_file" in file_entries
    assert "Mock_some_path_other_file" in file_entries


def test_filesystem_entry__iterdir_raises(
    exfat_fs: ExfatFilesystem,
    other_file: ExfatFilesystemEntry,
) -> None:
    with pytest.raises(NotADirectoryError):
        list(other_file._iterdir())


def test_filesystem_entry_iterdir(
    exfat_fs: ExfatFilesystem,
    some_path: ExfatFilesystemEntry,
) -> None:
    file_names = list(some_path.iterdir())

    assert len(file_names) == 2
    assert "some_file" in file_names
    assert "other_file" in file_names


def test_filesystem_entry_iterdir_raises(
    exfat_fs: ExfatFilesystem,
    other_file: ExfatFilesystemEntry,
) -> None:
    with pytest.raises(NotADirectoryError):
        list(other_file.iterdir())


def test_filesystem_entry_scandir(
    exfat_fs: ExfatFilesystem,
    some_path: ExfatFilesystemEntry,
) -> None:
    file_names = [entry.name for entry in some_path.scandir()]

    assert len(file_names) == 2
    assert "some_file" in file_names
    assert "other_file" in file_names


def test_filesystem_entry_scandir_raises(
    exfat_fs: ExfatFilesystem,
    other_file: ExfatFilesystemEntry,
) -> None:
    with pytest.raises(NotADirectoryError):
        list(other_file.scandir())


def test_filesystem_entry_is_symlink(
    exfat_fs: ExfatFilesystem,
    other_file: ExfatFilesystemEntry,
    some_path: ExfatFilesystemEntry,
) -> None:
    assert not other_file.is_symlink()
    assert not some_path.is_symlink()


def test_filesystem_entry_is_dir(
    exfat_fs: ExfatFilesystem,
    other_file: ExfatFilesystemEntry,
    some_path: ExfatFilesystemEntry,
) -> None:
    assert not other_file.is_dir()
    assert some_path.is_dir()
    assert some_path.is_dir(follow_symlinks=False)


def test_filesystem_entry_is_file(
    exfat_fs: ExfatFilesystem,
    other_file: ExfatFilesystemEntry,
    some_path: ExfatFilesystemEntry,
) -> None:
    assert other_file.is_file()
    assert other_file.is_file(follow_symlinks=False)
    assert not some_path.is_file()



from __future__ import annotations

from typing import TYPE_CHECKING, Callable
from unittest.mock import patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.vbox import VBoxLoader
from dissect.target.target import Target

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], tmp_path: Path) -> None:
    """Test that we correctly use ``VBoxLoader`` when opening a ``Target``."""
    path = tmp_path / "test.vbox"
    path.touch()

    with (
        patch("dissect.hypervisor.descriptor.vbox.VBox") as mock_vbox,
        patch("dissect.target.container.open"),
        patch("dissect.target.target.Target.apply"),
    ):
        mock_vbox.return_value = mock_vbox
        mock_vbox.disks.return_value = ["mock.vdi"]

        target = opener(path)
        assert isinstance(target._loader, VBoxLoader)
        assert target.path == path


def test_loader(tmp_path: Path) -> None:
    path = tmp_path / "test.vbox"
    path.touch()

    with (
        patch("dissect.hypervisor.descriptor.vbox.VBox") as mock_vbox,
        patch("dissect.target.container.open") as mock_container_open,
    ):
        mock_vbox.return_value = mock_vbox
        mock_vbox.disks.return_value = ["mock.vdi"]

        loader = loader_open(path)
        assert isinstance(loader, VBoxLoader)

        t = Target()
        loader.map(t)

        assert len(t.disks) == 1
        mock_container_open.assert_called_with(loader.base_path / "mock.vdi")



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dissect.target import container
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target

RE_VOLUME_ID = re.compile(r"(?:file=)?([^:]+):([^,]+)")


class ProxmoxLoader(Loader):
    """Loader for Proxmox VM configuration files.

    Proxmox uses volume identifiers in the format of ``storage_id:volume_id``. The ``storage_id`` maps to a
    storage configuration in ``/etc/pve/storage.cfg``. The ``volume_id`` is the name of the volume within
    that configuration.

    This loader currently does not support parsing the storage configuration, so it will attempt to open the
    volume directly from the same directory as the configuration file, or from ``/dev/pve/`` (default LVM config).
    If the volume is not found, it will log a warning.
    """

    @staticmethod
    def detect(path: Path) -> bool:
        if path.suffix.lower() != ".conf":
            return False

        with path.open("rb") as fh:
            lines = fh.read(512).split(b"\n")
            needles = [b"cpu:", b"memory:", b"name:"]
            return all(any(needle in line for line in lines) for needle in needles)

    def map(self, target: Target) -> None:
        with self.path.open("rt") as fh:
            for line in fh:
                if not (line := line.strip()):
                    continue

                key, value = line.split(":", 1)
                value = value.strip()

                # https://pve.proxmox.com/wiki/Storage
                if (
                    key.startswith(("scsi", "sata", "ide", "virtio"))
                    and key[-1].isdigit()
                    and (match := RE_VOLUME_ID.match(value))
                ):
                    storage_id, volume_id = match.groups()

                    # TODO: parse the storage information from /etc/pve/storage.cfg
                    # For now, let's try a few assumptions
                    disk_path = None
                    if (path := self.base_path.joinpath(volume_id)).exists() or (
                        path := self.base_path.joinpath("/dev/pve/").joinpath(volume_id)
                    ).exists():
                        disk_path = path

                    if disk_path:
                        try:
                            target.disks.add(container.open(disk_path))
                        except Exception:
                            target.log.exception("Failed to open disk: %s", disk_path)
                    else:
                        target.log.warning("Unable to find disk: %s:%s", storage_id, volume_id)



from __future__ import annotations

import argparse
from unittest.mock import patch

from dissect.target.plugin import PluginRegistry
from dissect.target.tools import build_pluginlist


def test_main_output() -> None:
    with (
        patch("argparse.ArgumentParser.parse_args", return_value=argparse.Namespace(verbose=0)),
        patch("dissect.target.tools.build_pluginlist.plugin.generate", return_value=PluginRegistry()),
        patch("builtins.print") as mock_print,
    ):
        build_pluginlist.main()

    expected_output = """
from dissect.target.plugin import (
    FailureDescriptor,
    FunctionDescriptor,
    FunctionDescriptorLookup,
    PluginDescriptor,
    PluginDescriptorLookup,
    PluginRegistry,
)

PLUGINS = PluginRegistry(__plugins__=PluginDescriptorLookup(__regular__={}, __os__={}, __child__={}), __functions__=FunctionDescriptorLookup(__regular__={}, __os__={}, __child__={}), __ostree__={}, __failed__=[])
"""  # noqa: E501

    mock_print.assert_called_with(expected_output)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

import dissect.target.helpers.hashutil as hashutil

if TYPE_CHECKING:
    from dissect.target.target import Target

HASHES = ("CAFEF00D" * 4, "F4CEF001" * 5, "DEADBEEF" * 8)


@pytest.fixture
def mock_target(target_win: Target) -> Target:
    target_win.fs.hash = lambda path: HASHES
    target_win.resolve = lambda path: Path(path)
    return target_win


def resolve_func(resolvable_path: str) -> str:
    return Path(f"/resolved{resolvable_path}")


def test_common() -> None:
    with Path(__file__).open("rb") as fh:
        output = hashutil.common(fh)

    assert len(output[0]) == 32
    assert len(output[1]) == 40
    assert len(output[2]) == 64



#!/usr/bin/env python
from __future__ import annotations

import argparse
import itertools
import logging
from typing import TYPE_CHECKING

from dissect.target.exceptions import (
    RegistryError,
    RegistryKeyNotFoundError,
    TargetError,
)
from dissect.target.target import Target
from dissect.target.tools.utils import (
    catch_sigpipe,
    configure_generic_arguments,
    process_generic_arguments,
)

if TYPE_CHECKING:
    from dissect.target.helpers.regutil import RegistryKey

log = logging.getLogger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    parser.add_argument("targets", metavar="TARGETS", nargs="+", help="Targets to load")
    parser.add_argument("-k", "--key", required=True, help="key to query")
    parser.add_argument("-kv", "--value", help="value to query")
    parser.add_argument("-d", "--depth", type=int, const=0, nargs="?", default=1, help="max depth of subkeys to print")
    parser.add_argument("-l", "--length", type=int, default=100, help="max length of key value to print")
    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()
    process_generic_arguments(args)

    try:
        for target in Target.open_all(args.targets):
            if not target.has_function("registry"):
                target.log.error("Target has no Windows Registry")
                continue

            try:
                keys = target.registry.keys(args.key)
                first_key = next(keys)

                print(target)

                for key in itertools.chain([first_key], keys):
                    try:
                        if args.value:
                            print(key.value(args.value))
                        else:
                            recursor(key, args.depth, 0, args.length)
                    except RegistryError:  # noqa: PERF203
                        log.exception("Failed to find registry value")

            except (RegistryKeyNotFoundError, StopIteration):
                target.log.error("Key %r does not exist", args.key)  # noqa: TRY400

            except Exception as e:
                target.log.error("Failed to iterate key: %s", e)  # noqa: TRY400
                target.log.debug("", exc_info=e)
    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    return 0


def recursor(key: RegistryKey, depth: int, indent: int, max_length: int = 100) -> None:
    class_name = ""
    if key.class_name:
        class_name = f" ({key.class_name})"

    print(" " * indent + f"+ {key.name!r} ({key.ts})" + class_name)

    for r in key.values():
        try:
            value = repr(r.value)
            if len(value) > max_length:
                value = value[:max_length] + "..."
            print(" " * indent + f"  - {r.name!r} {value}")
        except NotImplementedError:  # noqa: PERF203
            continue

    if depth == 0:
        return

    for subkey in key.subkeys():
        recursor(subkey, depth - 1, indent + 2, max_length)


if __name__ == "__main__":
    main()



from __future__ import annotations

import base64
import functools
import inspect
import os
from itertools import tee
from pathlib import Path
from types import GeneratorType
from typing import TYPE_CHECKING, Any, Callable

from flow.record import RecordReader, RecordWriter
from flow.record.base import HAS_ZSTD

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flow.record.adapter.stream import StreamReader, StreamWriter

    from dissect.target.target import Target

Tee = type(tee([], 1)[0])

IGNORE_CACHE = os.getenv("IGNORE_CACHE", "0") == "1"
ONLY_READ_CACHE = os.getenv("ONLY_READ_CACHE", "0") == "1"
REWRITE_CACHE = os.getenv("REWRITE_CACHE", "0") == "1"


class LineWriter:
    def __init__(self, path: Path):
        self.fh = path.open("w")

    def write(self, line: str) -> None:
        self.fh.write(line + "\n")

    def close(self) -> None:
        self.fh.close()


class LineReader:
    def __init__(self, path: Path):
        self.path = path

    def __iter__(self) -> Iterator[str]:
        with self.path.open() as fh:
            for line in fh:
                yield line.strip()


class CacheWriter:
    def __init__(self, path: Path, temp: Path, reader: Iterator[Any], writer: StreamWriter | LineWriter):
        self.path = path
        self.temp = temp
        self.reader = reader
        self.writer = writer

    def __iter__(self) -> Iterator[Any]:
        for obj in self.reader:
            self.writer.write(obj)
            yield obj
        self.close()

    def close(self) -> None:
        self.writer.close()
        try:
            self.temp.rename(self.path)
        except OSError:
            pass


class Cache:
    def __init__(self, func: Callable, no_cache: bool = False, cls: type | None = None):
        self.func = func
        self.no_cache = no_cache

        module = inspect.getmodule(cls or func)
        module = f"{module.__name__}." if module else ""
        qualname = f"{cls.__name__}.{func.__name__}" if cls else func.__qualname__

        self.fname = f"{module}{qualname}"
        self.wrapper = None

    def open_reader(self, path: Path, output: str) -> StreamReader | LineReader | None:
        if output == "record":
            return RecordReader(str(path))
        if output == "yield":
            return LineReader(path)
        return None

    def open_writer(self, path: Path, output: str) -> StreamWriter | LineWriter | None:
        if output == "record":
            return RecordWriter(str(path))
        if output == "yield":
            return LineWriter(path)
        return None

    def cache_path(self, target: Target, key: tuple) -> Path | None:
        cache_dir = getattr(target._config, "CACHE_DIR", None) if target._config else None
        if not cache_dir:
            return None

        path_key = base64.b64encode(repr(key).encode()).decode("utf8")
        ext = "zstd" if HAS_ZSTD else "rec"
        fname = f"{self.fname}.{path_key}.{ext}"
        return Path(cache_dir).joinpath(Path(target.path).name, fname)

    def call(self, *args, **kwargs) -> Any:
        target: Target = args[0].target

        output = getattr(self.wrapper, "__output__", None)
        if output not in ("record", "yield", "none"):
            # Cache property and default outputs on the target object itself
            func_cache = target._cache.setdefault(self.fname, {})
            key = (args[1:], frozenset(sorted(kwargs.items())))

            if key not in func_cache:
                func_cache[key] = self.func(*args, **kwargs)

            if isinstance(func_cache[key], (GeneratorType, Tee)):
                func_cache[key], cache_result = tee(func_cache[key])
                return cache_result

            return func_cache[key]

        key = (args[1:], sorted(kwargs.items()))
        cache_file = self.cache_path(target, key)

        # The default policy is READ cache if available else WRITE it (reading
        # takes precedense over writing) and only if the cache directory is
        # configured in the environment. Furthermore, reading is disabled if
        # there is no cache file for the specific function called.
        #
        # Further the IGNORE_CACHE takes precedence over ONLY_READ_CACHE takes
        # precedence over REWRITE_CACHE. So setting an option of a lower
        # precedence to True has no effect if a higher precedence option is
        # already set to True.
        #
        # This set of rules makes sure the 'safest' cache option that is set
        # will be used.
        read_file_cache = True
        write_file_cache = True

        if cache_file:
            if not cache_file.exists():
                read_file_cache = False
        else:
            read_file_cache = False
            write_file_cache = False

        if IGNORE_CACHE:
            read_file_cache = False
            write_file_cache = False
        elif ONLY_READ_CACHE:
            write_file_cache = False
        elif REWRITE_CACHE:
            read_file_cache = False

        if read_file_cache:
            target.log.debug("Reading from cache file: %s", cache_file)
            if os.access(cache_file, os.R_OK, effective_ids=bool(os.supports_effective_ids)):
                if cache_file.stat().st_size != 0:
                    try:
                        reader = self.open_reader(cache_file, output)
                    except Exception as e:
                        target.log.warning("Cache will NOT be used. Error opening cache file: %s", cache_file)
                        target.log.debug("", exc_info=e)
                    else:
                        target.log.info("Using cache for function: %s", self.fname)
                        return reader
                else:
                    target.log.warning("Cache will NOT be used. File is empty: %s", cache_file)
            else:
                target.log.warning("Cache will NOT be used. No permissions to read cache file: %s", cache_file)
        elif write_file_cache:
            dir_mode = getattr(target._config, "CACHE_DIR_MODE", 0o777) if target._config else 0o777
            file_mode = getattr(target._config, "CACHE_FILE_MODE", 0o666) if target._config else 0o666

            temp_dir = cache_file.parent
            temp_path = cache_file.with_name(f"_{cache_file.name}")

            if not temp_dir.exists():
                try:
                    temp_dir.mkdir(mode=dir_mode, parents=True)
                except Exception as e:
                    target.log.warning(
                        "Cache will NOT be written. Unable to create cache directory: %s (%s)", temp_dir, e
                    )

            if os.access(temp_dir, os.W_OK | os.R_OK | os.X_OK, effective_ids=bool(os.supports_effective_ids)):
                if temp_path.exists():
                    try:
                        temp_path.unlink()
                    except Exception as e:
                        target.log.warning(
                            "Cache will NOT be written. Unable to remove pre-existing cache temp file: %s (%s)",
                            temp_path,
                            e,
                        )

                if not temp_path.exists():
                    try:
                        writer = self.open_writer(temp_path, output)
                        try:
                            # Set permissions
                            temp_path.chmod(file_mode)
                        except OSError as e:
                            target.log.debug(
                                "Setting permissions on temp cache file failed, "
                                "continuing with existing permissions: %s (%s)",
                                temp_path,
                                e,
                            )
                        target.log.debug("Caching to file: %s", temp_path)
                        return CacheWriter(cache_file, temp_path, self.func(*args, **kwargs), writer)
                    except Exception as e:
                        target.log.error("Cache will NOT be written. Failed to cache to file: %s (%s)", cache_file, e)  # noqa: TRY400
                        target.log.debug("", exc_info=e)
                        try:
                            temp_path.unlink()
                        except Exception as e:
                            target.log.warning("Unable to remove cache temp file: %s (%s)", temp_path, e)
            else:
                target.log.warning(
                    "Cache will NOT be written. No permissions to write cache file in directory: %s",
                    temp_dir,
                )

        return self.func(*args, **kwargs)


def wrap(func: Callable, no_cache: bool = False, cls: type | None = None) -> Callable:
    cache = Cache(func, no_cache=no_cache, cls=cls)

    @functools.wraps(func)
    def cache_wrapper(*args, **kwargs) -> Any:
        if cache.no_cache:
            return cache.func(*args, **kwargs)
        return cache.call(*args, **kwargs)

    cache.wrapper = cache_wrapper
    return cache_wrapper



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugin import OperatingSystem
from dissect.target.plugins.os.unix.bsd.openbsd._os import OpenBsdPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_bsd_openbsd_os_detection(target_bare: Target) -> None:
    """Test if we detect OpenBSD correctly."""

    fs = VirtualFilesystem()
    fs.map_file_fh("/etc/myname", BytesIO(b"hostname"))
    fs.makedirs("/bsd")

    target_bare.filesystems.add(fs)
    target_bare.apply()

    assert OpenBsdPlugin.detect(target_bare)
    assert isinstance(target_bare._os, OpenBsdPlugin)
    assert target_bare.os == OperatingSystem.BSD



from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING, BinaryIO

from dissect.fve import luks

from dissect.target.exceptions import VolumeSystemError
from dissect.target.helpers.keychain import KeyType
from dissect.target.volume import EncryptedVolumeSystem, Volume

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.util.stream import AlignedStream

log = logging.getLogger(__name__)


class LUKSVolumeSystemError(VolumeSystemError):
    pass


class LUKSVolumeSystem(EncryptedVolumeSystem):
    __type__ = "luks"

    def __init__(self, fh: BinaryIO | list[BinaryIO], *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.luks = luks.LUKS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return luks.is_luks_volume(fh)

    def _volumes(self) -> Iterator[Volume]:
        if isinstance(self.fh, Volume):
            volume_details = {
                "number": self.fh.number,
                "offset": self.fh.offset,
                "vtype": self.fh.type,
                "name": self.fh.name,
            }
        else:
            volume_details = {
                "number": 1,
                "offset": 0,
                "vtype": None,
                "name": None,
            }

        stream = self.unlock_volume()
        yield Volume(
            fh=stream,
            size=stream.size,
            raw=self.fh,
            disk=self.disk,
            vs=self,
            **volume_details,
        )

    def unlock_with_volume_encryption_key(
        self, key: bytes, keyslot: int | None = None, is_wildcard: bool = False
    ) -> None:
        try:
            if keyslot is None:
                for keyslot in self.luks.keyslots:
                    try:
                        self.luks.unlock(key, keyslot)
                        break
                    except ValueError:
                        continue
                else:
                    raise ValueError("Failed to find matching keyslot for provided volume encryption key")  # noqa: TRY301
            else:
                self.luks.unlock(key, keyslot)

            log.debug("Unlocked LUKS volume with provided volume encryption key")
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock LUKS volume with provided volume encryption key")

    def unlock_with_passphrase(self, passphrase: str, keyslot: int | None = None, is_wildcard: bool = False) -> None:
        try:
            self.luks.unlock_with_passphrase(passphrase, keyslot)
            log.debug("Unlocked LUKS volume with provided passphrase")
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock LUKS volume with provided passphrase")

    def unlock_with_key_file(
        self, key_file: pathlib.Path, keyslot: int | None = None, is_wildcard: bool = False
    ) -> None:
        if not key_file.exists():
            if not is_wildcard:
                log.error("Provided key file does not exist: %s", key_file)
            return

        try:
            self.luks.unlock_with_key_file(key_file, keyslot=keyslot)
            log.debug("Unlocked LUKS volume with key file %s", key_file)
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock LUKS volume with key file %s", key_file)

    def unlock_volume(self) -> AlignedStream:
        keyslots = list(map(str, self.luks.keyslots.keys()))
        keys = self.get_keys_for_identifiers(keyslots) + self.get_keys_without_identifier()

        for key in keys:
            try:
                keyslot = int(key.identifier)
            except Exception:
                keyslot = None

            if key.key_type == KeyType.RAW:
                self.unlock_with_volume_encryption_key(key.value, keyslot, key.is_wildcard)
            if key.key_type == KeyType.PASSPHRASE:
                self.unlock_with_passphrase(key.value, keyslot, key.is_wildcard)
            elif key.key_type == KeyType.FILE:
                key_file = pathlib.Path(key.value)
                self.unlock_with_key_file(key_file, keyslot, key.is_wildcard)

            if self.luks.unlocked:
                log.info("Volume %s unlocked with %s (keyslot: %d)", self.fh, key, self.luks._active_keyslot_id)
                break

        if self.luks.unlocked:
            return self.luks.open()
        raise LUKSVolumeSystemError("Failed to unlock LUKS volume")



from __future__ import annotations

from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime

from dissect.target.plugins.os.unix.bsd.darwin.macos.user import UserPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_unix_bsd_darwin_macos_user(target_macos_users: Target) -> None:
    target_macos_users.add_plugin(UserPlugin)

    results = list(target_macos_users.account_policy())
    assert len(results) == 2

    assert results[0].creation_time == datetime("2022-09-19T15:15:34.564508Z")
    assert results[0].password_last_time is None
    assert results[0].failed_login_time is None
    assert results[0].failed_login_count is None
    assert results[0].generateduid == "BD6AC542-F7BE-1337-B2DB-30F9EE37E133"
    assert results[0].source == "/var/db/dslocal/nodes/Default/users/_dissect.plist"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.loaders.pvs import PvsLoader

if TYPE_CHECKING:
    from pathlib import Path


class PvmLoader(PvsLoader):
    """Parallels VM directory (.pvm)."""

    def __init__(self, path: Path, **kwargs):
        super().__init__(path.joinpath("config.pvs"), **kwargs)

    @staticmethod
    def detect(path: Path) -> bool:
        return path.is_dir() and path.suffix.lower() == ".pvm" and path.joinpath("config.pvs").exists()



from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.evidence import AsdfSnapshot, AsdfStream
from dissect.evidence.asdf import FILE_MAGIC

from dissect.target.container import Container

if TYPE_CHECKING:
    from pathlib import Path


class AsdfContainer(Container):
    __type__ = "asdf"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        file_container = fh

        if not hasattr(file_container, "read"):
            file_container = AsdfSnapshot(fh.open("rb"))

        if isinstance(file_container, AsdfSnapshot):
            file_container = fh.open(0)

        if not isinstance(file_container, AsdfStream):
            raise TypeError(f"Invalid type for AsdfContainer: {fh}")

        self.asdf = file_container

        super().__init__(fh, self.asdf.size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return fh.read(4) == FILE_MAGIC

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        return path.suffix.lower() == ".asdf"

    def read(self, length: int) -> bytes:
        return self.asdf.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.asdf.seek(offset, whence)

    def tell(self) -> int:
        return self.asdf.tell()

    def close(self) -> None:
        self.asdf.close()



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.locate.gnulocate import (
    GNULocatePlugin,
    GNULocateRecord,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_gnulocate(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file("/var/cache/locate/locatedb", absolute_path("_data/plugins/os/unix/locate/locatedb"))
    target_unix.add_plugin(GNULocatePlugin)

    records = list(target_unix.gnulocate.locate())

    assert len(records) == 3575
    assert isinstance(records[0], type(GNULocateRecord()))

    assert records[0].path.as_posix() == "/"
    assert records[1].path.as_posix() == "/.dockerenv"

    # test namespace plugin
    records = list(target_unix.locate.locate())

    assert len(records) == 3575
    assert isinstance(records[0], type(GNULocateRecord()))

    assert records[0].path.as_posix() == "/"
    assert records[1].path.as_posix() == "/.dockerenv"



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

NETSTAT_HEADER = f"Active Internet connections (only servers)\n{'Proto':<10}{'Recv-Q':^10}{'Send-Q':^10}{'Local Address':^20}{'Foreign Address':^20}{'State':^10}{'User':^15}{'Inode':^10}{'PID/Program name':^10}{'Command':>10}"  # noqa
NETSTAT_TEMPLATE = "{protocol:<12}{receive_queue:<10}{transmit_queue:<11}{local_addr:<19}{remote_addr:<20}{state:<13}{owner:<12}{inode:<8}{pid_program:<19}{cmdline}"  # noqa


class NetstatPlugin(Plugin):
    """Linux volatile netstat plugin."""

    def check_compatible(self) -> None:
        if not self.target.has_function("proc"):
            raise UnsupportedPluginError("proc filesystem not available")

    @export(output="yield")
    def netstat(self) -> Iterator[str]:
        """This plugin mimics the output `netstat -tunelwap` would generate on a Linux machine."""
        sockets = chain(
            self.target.sockets.tcp(),
            self.target.sockets.udp(),
            self.target.sockets.raw(),
        )

        yield NETSTAT_HEADER

        for record in sockets:
            local_addr = f"{record.local_ip}:{record.local_port}"
            remote_addr = f"{record.remote_ip}:{record.remote_port}"
            pid_program = f"{record.pid}/{record.name}"

            yield NETSTAT_TEMPLATE.format(
                protocol=record.protocol,
                receive_queue=record.rx_queue,
                transmit_queue=record.tx_queue,
                local_addr=local_addr,
                remote_addr=remote_addr,
                state=record.state,
                owner=record.owner,
                inode=record.inode,
                pid_program=pid_program,
                cmdline=record.cmdline,
            )



from __future__ import annotations

from enum import IntEnum


class IANAProtocol(IntEnum):
    """Enum of IANA Protocol numbers.

    Resources:
        - https://www.iana.org/assignments/protocol-numbers/protocol-numbers.xhtml
    """

    HOPOPT = 0
    ICMP = 1
    IGMP = 2
    GGP = 3
    IPV4 = 4
    ST = 5
    TCP = 6
    CBT = 7
    EGP = 8
    IGP = 9
    BBN_RCC_MON = 10
    NVP_II = 11
    PUP = 12
    ARGUS = 13
    EMCON = 14
    XNET = 15
    CHAOS = 16
    UDP = 17
    MUX = 18
    DCN_MEAS = 19
    HMP = 20
    PRM = 21
    XNS_IDP = 22
    TRUNK_1 = 23
    TRUNK_2 = 24
    LEAF_1 = 25
    LEAF_2 = 26
    RDP = 27
    IRTP = 28
    ISO_TP4 = 29
    NETBLT = 30
    MFE_NSP = 31
    MERIT_INP = 32
    DCCP = 33
    _3PC = 34
    IDPR = 35
    XTP = 36
    DDP = 37
    IDPR_CMTP = 38
    TPPP = 39
    IL = 40
    IPV6 = 41
    SDRP = 42
    IPV6_ROUTE = 43
    IPV6_FRAG = 44
    IDRP = 45
    RSVP = 46
    GRE = 47
    DSR = 48
    BNA = 49
    ESP = 50
    AH = 51
    I_NLSP = 52
    SWIPE = 53
    NARP = 54
    MIN_IPV4 = 55
    TLSP = 56
    SKIP = 57
    IPV6_ICMP = 58
    IPV6_NONXT = 59
    IPV6_OPTS = 60
    ANY_HOST_INTERNAL_PROTOCOL = 61
    CFTP = 62
    ANY_LOCAL_NETWORK = 63
    SAT_EXPAK = 64
    KRYPTOLAN = 65
    RVD = 66
    IPPC = 67
    ANY_DISTRIBUTED_FILE_SYSTEM = 68
    SAT_MON = 69
    VISA = 70
    IPCV = 71
    CPNX = 72
    CPHB = 73
    WSN = 74
    PVP = 75
    BR_SAT_MON = 76
    SUN_ND = 77
    WB_MON = 78
    WB_EXPAK = 79
    ISO_IP = 80
    VMTP = 81
    SECURE_VMTP = 82
    VINES = 83
    IPTM = 84
    NSFNET_IGP = 85
    DGP = 86
    TCF = 87
    EIGRP = 88
    OSPFIGP = 89
    SPRITE_RPC = 90
    LARP = 91
    MTP = 92
    AX_25 = 93
    IPIP = 94
    MICP = 95
    SCC_SP = 96
    ETHERIP = 97
    ENCAP = 98
    ANY_PRIVATE_ENCRYPTION_SCHEME = 99
    GMTP = 100
    IFMP = 101
    PNNI = 102
    PIM = 103
    ARIS = 104
    SCPS = 105
    QNX = 106
    AN = 107
    IPCOMP = 108
    SNP = 109
    COMPAQ_PEER = 110
    IPX_IN_IP = 111
    VRRP = 112
    PGM = 113
    ANY_0_HOP_PROTOCOL = 114
    L2TP = 115
    DDX = 116
    IATP = 117
    STP = 118
    SRP = 119
    UTI = 120
    SMP = 121
    SM = 122
    PTP = 123
    ISIS_OVER_IPV4 = 124
    FIRE = 125
    CRTP = 126
    CRUDP = 127
    SSCOPMCE = 128
    IPLT = 129
    SPS = 130
    PIPE = 131
    SCTP = 132
    FC = 133
    RSVP_E2E_IGNORE = 134
    MOBILITY_HEADER = 135
    UDP_LITE = 136
    MPLS_IN_IP = 137
    MANET = 138
    HIP = 139
    SHIM6 = 140
    WESP = 141
    ROHC = 142
    ETHERNET = 143
    AGGFRAG = 144
    NSH = 145
    HOMA = 146
    BIT_EMU = 147
    # Unassigned = 148-252
    # Use for experimentation and testing = 253-254
    RESERVED_255 = 255

    @classmethod
    def _missing_(cls, value: int) -> IANAProtocol:
        # Allows values other than defined members
        member = int.__new__(cls, value)

        if value >= 148 and value <= 252:
            member._name_ = f"UNASSIGNED_{value!s}"
        elif value in (253, 254):
            member._name_ = f"USE_FOR_EXPERIMENTATION_AND_TESTING_{value!s}"
        else:
            member._name_ = f"UNKNOWN_{value!s}"

        member._value_ = value
        return member



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.browser.browser import (
    GENERIC_COOKIE_FIELDS,
    GENERIC_DOWNLOAD_RECORD_FIELDS,
    GENERIC_EXTENSION_RECORD_FIELDS,
    GENERIC_HISTORY_RECORD_FIELDS,
    GENERIC_PASSWORD_RECORD_FIELDS,
    BrowserPlugin,
)
from dissect.target.plugins.apps.browser.chromium import (
    CHROMIUM_DOWNLOAD_RECORD_FIELDS,
    ChromiumMixin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


class BravePlugin(ChromiumMixin, BrowserPlugin):
    """Brave browser plugin."""

    __namespace__ = "brave"

    DIRS = (
        # Windows
        "AppData/Local/BraveSoftware/Brave-Browser/User Data/Default",
        "AppData/Local/BraveSoftware/Brave-Browser/User Data/Profile*",
        "AppData/Roaming/BraveSoftware/Brave-Browser/User Data/Default",
        "AppData/Roaming/BraveSoftware/Brave-Browser/User Data/Profile*",
        # Linux
        ".config/BraveSoftware/Default",
        ".config/BraveSoftware/Profile*",
        ".config/BraveSoftware/Brave-Browser/Default",
        ".config/BraveSoftware/Brave-Browser/Profile*",
        "snap/brave/*/.config/BraveSoftware/Brave-Browser/Default",
        "snap/brave/*/.config/BraveSoftware/Brave-Browser/Profile*",
        # Macos
        "Library/Application Support/BraveSoftware/Default",
        "Library/Application Support/BraveSoftware/Brave-Browser/Default",
        "Library/Application Support/BraveSoftware/Profile*",
        "Library/Application Support/BraveSoftware/Brave-Browser/Profile*",
    )

    BrowserHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/brave/history", GENERIC_HISTORY_RECORD_FIELDS
    )

    BrowserCookieRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/brave/cookie", GENERIC_COOKIE_FIELDS
    )

    BrowserDownloadRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/brave/download", GENERIC_DOWNLOAD_RECORD_FIELDS + CHROMIUM_DOWNLOAD_RECORD_FIELDS
    )

    BrowserExtensionRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/brave/extension", GENERIC_EXTENSION_RECORD_FIELDS
    )

    BrowserPasswordRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/brave/password", GENERIC_PASSWORD_RECORD_FIELDS
    )

    @export(record=BrowserHistoryRecord)
    def history(self) -> Iterator[BrowserHistoryRecord]:
        """Return browser history records for Brave."""
        yield from super().history("brave")

    @export(record=BrowserCookieRecord)
    def cookies(self) -> Iterator[BrowserCookieRecord]:
        """Return browser cookie records for Brave."""
        yield from super().cookies("brave")

    @export(record=BrowserDownloadRecord)
    def downloads(self) -> Iterator[BrowserDownloadRecord]:
        """Return browser download records for Brave."""
        yield from super().downloads("brave")

    @export(record=BrowserExtensionRecord)
    def extensions(self) -> Iterator[BrowserExtensionRecord]:
        """Return browser extension records for Brave."""
        yield from super().extensions("brave")

    @export(record=BrowserPasswordRecord)
    def passwords(self) -> Iterator[BrowserPasswordRecord]:
        """Return browser password records for Brave."""
        yield from super().passwords("brave")



from __future__ import annotations

from unittest.mock import Mock, create_autospec

import pytest

from dissect.target.filesystems.fat import (
    FatFilesystem,
    FileNotFoundError,
    NotADirectoryError,
    fat_exc,
)


@pytest.mark.parametrize(
    ("raised_exception", "expected_exception"),
    [
        (fat_exc.FileNotFoundError, FileNotFoundError),
        (fat_exc.Error, FileNotFoundError),
        (fat_exc.NotADirectoryError, NotADirectoryError),
    ],
)
def test_get_entry(raised_exception: Exception, expected_exception: Exception) -> None:
    """Test whether the raised exception generates the expected exception."""
    mocked_fs = create_autospec(FatFilesystem)
    mocked_fs.fatfs = Mock()
    mocked_fs._get_entry = FatFilesystem._get_entry

    mocked_fs.fatfs.get.side_effect = [raised_exception]

    with pytest.raises(expected_exception):
        mocked_fs._get_entry(mocked_fs, path="")



from __future__ import annotations

import logging
import re
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Callable

import pytest
from flow.record import GroupedRecord

from dissect.target.plugins.os.windows.tasks._plugin import TaskRecord, TasksPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


@pytest.fixture
def setup_tasks_test(target_win: Target, fs_win: Filesystem) -> None:
    xml_task_file = absolute_path("_data/plugins/os/windows/tasks/MapsToastTask")
    atjob_task_file = absolute_path("_data/plugins/os/windows/tasks/AtTask.job")

    fs_win.map_file("windows/system32/tasks/Microsoft/Windows/Maps/MapsToastTask", xml_task_file)
    fs_win.map_file(
        "windows/system32/GroupPolicy/DataStore/ANY_SID/Machine/Preferences/ScheduledTasks/test_xml.xml", xml_task_file
    )
    fs_win.map_file("windows/tasks/AtTask.job", atjob_task_file)

    target_win.add_plugin(TasksPlugin)


@pytest.fixture
def setup_invalid_tasks_test(target_win: Target, fs_win: Filesystem, setup_tasks_test: None) -> None:
    xml_task_file_invalid = absolute_path("_data/plugins/os/windows/tasks/InvalidTask")

    fs_win.map_file("windows/system32/tasks/Microsoft/Windows/Maps/InvalidTask", xml_task_file_invalid)
    fs_win.map_file(
        "windows/system32/GroupPolicy/DataStore/ANY_SID/Machine/Preferences/ScheduledTasks/invalid_xml.xml",
        xml_task_file_invalid,
    )

    target_win.add_plugin(TasksPlugin)


def assert_xml_task_properties(xml_task: TaskRecord) -> None:
    assert str(xml_task.uri) == "\\Microsoft\\Windows\\Maps\\MapsToastTask"
    assert (
        xml_task.security_descriptor
        == "D:(A;;0x111FFFFF;;;SY)(A;;0x111FFFFF;;;BA)(A;;0x111FFFFF;;;S-1-5-80-3028837079-3186095147-955107200-3701964851-1150726376)(A;;FRFX;;;AU)"  # noqa: E501
    )
    assert xml_task.source is None
    assert xml_task.date == datetime(2014, 11, 5, 0, 0, 0, tzinfo=timezone.utc)
    assert xml_task.last_run_date is None
    assert xml_task.author == "$(@%SystemRoot%\\system32\\mapstoasttask.dll,-600)"
    assert xml_task.version is None
    assert xml_task.description == "$(@%SystemRoot%\\system32\\mapstoasttask.dll,-602)"
    assert xml_task.documentation is None
    assert xml_task.principal_id == "Users"
    assert xml_task.user_id is None
    assert xml_task.logon_type is None
    assert xml_task.group_id == "S-1-5-4"
    assert xml_task.display_name == "test_xml.xml"
    assert xml_task.run_level is None
    assert xml_task.process_token_sid_type is None
    assert xml_task.required_privileges is None
    assert xml_task.allow_start_on_demand is None
    assert xml_task.restart_on_failure_interval is None
    assert xml_task.restart_on_failure_count is None
    assert xml_task.mutiple_instances_policy == "Queue"
    assert not xml_task.disallow_start_on_batteries
    assert not xml_task.stop_going_on_batteries
    assert xml_task.start_when_available
    assert xml_task.network_profile_name is None
    assert xml_task.run_only_network_available is None
    assert xml_task.wake_to_run is None
    assert xml_task.enabled
    assert xml_task.hidden
    assert xml_task.delete_expired_task_after is None
    assert xml_task.idle_duration is None
    assert xml_task.idle_wait_timeout is None
    assert not xml_task.idle_stop_on_idle_end
    assert not xml_task.idle_restart_on_idle
    assert xml_task.network_settings_name is None
    assert xml_task.network_settings_id is None
    assert xml_task.execution_time_limit == "PT5S"
    assert xml_task.priority is None
    assert xml_task.run_only_idle is None
    assert xml_task.unified_scheduling_engine
    assert xml_task.disallow_start_on_remote_app_session is None
    assert xml_task.data is None


def assert_at_task_properties(at_task: TaskRecord) -> None:
    assert at_task.uri is None
    assert at_task.security_descriptor is None
    assert str(at_task.task_path) == "c:\\Windows\\tasks\\AtTask.job"
    assert at_task.date is None
    assert at_task.last_run_date == datetime(2023, 5, 21, 10, 44, 25, 794000, tzinfo=timezone.utc)
    assert at_task.author == "user1"
    assert at_task.version == "1"
    assert at_task.description == "At job task for testing purposes"
    assert at_task.documentation is None
    assert at_task.principal_id is None
    assert at_task.user_id is None
    assert at_task.logon_type is None
    assert at_task.group_id is None
    assert at_task.display_name is None
    assert at_task.run_level is None
    assert at_task.process_token_sid_type is None
    assert at_task.required_privileges is None
    assert at_task.allow_start_on_demand is None
    assert at_task.restart_on_failure_interval is None
    assert at_task.restart_on_failure_count == "0"
    assert at_task.mutiple_instances_policy is None
    assert at_task.disallow_start_on_batteries
    assert at_task.stop_going_on_batteries
    assert at_task.start_when_available is None
    assert at_task.network_profile_name is None
    assert not at_task.run_only_network_available
    assert at_task.wake_to_run
    assert at_task.enabled
    assert not at_task.hidden
    assert at_task.delete_expired_task_after is None
    assert at_task.idle_duration == "PT15M"
    assert at_task.idle_wait_timeout == "PT1H"
    assert at_task.idle_stop_on_idle_end
    assert not at_task.idle_restart_on_idle
    assert at_task.network_settings_name is None
    assert at_task.network_settings_id is None
    assert at_task.execution_time_limit == "P3D"
    assert at_task.priority == "normal"
    assert at_task.run_only_idle
    assert at_task.unified_scheduling_engine is None
    assert at_task.disallow_start_on_remote_app_session is None
    assert at_task.data == "[]"


def assert_xml_task_grouped_properties(xml_task_grouped: GroupedRecord) -> None:
    assert xml_task_grouped.action_type == "ComHandler"
    assert xml_task_grouped.class_id == "{9885AEF2-BD9F-41E0-B15E-B3141395E803}"
    assert xml_task_grouped.com_data == "<Data>$(Arg0);$(Arg1);$(Arg2);$(Arg3);$(Arg4);$(Arg5);$(Arg6);$(Arg7)</Data>"
    assert xml_task_grouped.data is None


def assert_at_task_grouped_exec(at_task_grouped: GroupedRecord) -> None:
    assert at_task_grouped.action_type == "Exec"
    assert at_task_grouped.arguments == ""
    assert at_task_grouped.command == "C:\\WINDOWS\\NOTEPAD.EXE"
    assert at_task_grouped.working_directory == "C:\\Documents and Settings\\John"


def assert_at_task_grouped_daily(at_task_grouped: GroupedRecord) -> None:
    assert at_task_grouped.days_between_triggers == 3
    assert at_task_grouped.end_boundary == datetime.fromisoformat("2023-05-12 00:00:00+00:00")
    assert at_task_grouped.execution_time_limit == "P3D"
    assert at_task_grouped.repetition_duration == "PT13H15M"
    assert at_task_grouped.repetition_interval == "PT12M"
    assert at_task_grouped.repetition_stop_duration_end
    assert at_task_grouped.start_boundary == datetime.fromisoformat("2023-05-11 00:00:00+00:00")
    assert_at_task_grouped_padding(at_task_grouped)


def assert_at_task_grouped_padding(at_task_grouped: GroupedRecord) -> None:
    assert at_task_grouped.padding == 0
    assert at_task_grouped.reserved2 == 0
    assert at_task_grouped.reserved3 == 0


def assert_at_task_grouped_monthlydow(at_task_grouped: GroupedRecord) -> None:
    assert at_task_grouped.records[0].enabled
    assert at_task_grouped.records[1].trigger_enabled
    assert at_task_grouped.start_boundary == datetime.fromisoformat("2023-05-11 00:00:00+00:00")
    assert at_task_grouped.end_boundary == datetime.fromisoformat("2023-05-20 00:00:00+00:00")
    assert at_task_grouped.repetition_interval == "PT1M"
    assert at_task_grouped.repetition_duration == "PT12H13M"
    assert at_task_grouped.repetition_stop_duration_end
    assert at_task_grouped.execution_time_limit == "P3D"
    assert at_task_grouped.which_week == [2]
    assert at_task_grouped.days_of_week == ["Wednesday"]
    assert at_task_grouped.months_of_year == ["June", "September"]
    assert_at_task_grouped_padding(at_task_grouped)


def assert_at_task_grouped_weekly(at_task_grouped: GroupedRecord) -> None:
    assert at_task_grouped.records[0].enabled
    assert at_task_grouped.records[1].trigger_enabled
    assert at_task_grouped.end_boundary == datetime.fromisoformat("2023-05-27 00:00:00+00:00")
    assert at_task_grouped.execution_time_limit == "P3D"
    assert at_task_grouped.repetition_duration == "PT1H"
    assert at_task_grouped.repetition_interval == "PT10M"
    assert at_task_grouped.repetition_stop_duration_end
    assert at_task_grouped.start_boundary == datetime.fromisoformat("2023-05-23 00:00:00+00:00")
    assert at_task_grouped.days_of_week == ["Monday", "Wednesday", "Friday"]
    assert at_task_grouped.unused == [0]
    assert at_task_grouped.weeks_between_triggers == 1
    assert_at_task_grouped_padding(at_task_grouped)


def assert_at_task_grouped_monthly_date(at_task_grouped: GroupedRecord) -> None:
    assert at_task_grouped.records[0].enabled
    assert at_task_grouped.records[1].trigger_enabled
    assert at_task_grouped.day_of_month == [15]
    assert at_task_grouped.months_of_year == ["March", "May", "June", "July", "August", "October"]
    assert at_task_grouped.end_boundary == datetime.fromisoformat("2023-05-29 00:00:00+00:00")
    assert at_task_grouped.execution_time_limit == "P3D"
    assert at_task_grouped.repetition_duration == "PT4H44M"
    assert at_task_grouped.repetition_interval == "PT17M"
    assert at_task_grouped.repetition_stop_duration_end
    assert at_task_grouped.start_boundary == datetime.fromisoformat("2023-05-23 00:00:00+00:00")


def assert_xml_task_trigger_properties(xml_task: GroupedRecord) -> None:
    assert xml_task.records[0].enabled
    assert xml_task.records[1].trigger_enabled
    assert xml_task.days_between_triggers == 1
    assert xml_task.start_boundary == datetime.fromisoformat("2023-05-12 00:00:00+00:00")


@pytest.mark.parametrize(
    ("assert_func", "marker"),
    [
        (assert_xml_task_properties, "test_xml.xml.*ComHandler"),
        (assert_xml_task_properties, "MapsToastTask.*toast"),
        (assert_at_task_properties, "AtTask"),
    ],
)
def test_single_record_properties(
    target_win: Target, setup_tasks_test: None, assert_func: Callable, marker: str
) -> None:
    records = list(target_win.tasks(group=True))
    assert len(records) == 18
    pat = re.compile(rf"{marker}")
    records = filter(lambda x: re.findall(pat, str(x)), records)
    assert_func(next(iter(records)))


@pytest.mark.parametrize(
    ("assert_func", "marker"),
    [
        (assert_xml_task_grouped_properties, "test_xml.xml.*ComHandler"),
        (assert_xml_task_grouped_properties, "MapsToastTask.*ComHandler"),
        (assert_xml_task_trigger_properties, "MapsToastTask.*trigger_enabled"),
        (assert_at_task_grouped_exec, "NOTEPAD.EXE"),
        (assert_at_task_grouped_daily, "PT13H15M"),
        (assert_at_task_grouped_monthlydow, "June"),
        (assert_at_task_grouped_weekly, "Friday"),
        (assert_at_task_grouped_monthly_date, "2023-05-29"),
    ],
)
def test_grouped_record_properties(
    target_win: Target, setup_invalid_tasks_test: pytest.fixture, assert_func: Callable, marker: str
) -> None:
    records = list(target_win.tasks(group=True))
    assert len(records) == 18
    pat = re.compile(rf"{marker}")
    grouped_records = filter(lambda x: re.findall(pat, str(x)) and isinstance(x, GroupedRecord), records)
    assert_func(next(iter(grouped_records)))


def test_xml_task_invalid(
    target_win: Target, setup_invalid_tasks_test: pytest.fixture, caplog: pytest.LogCaptureFixture
) -> None:
    caplog.clear()
    with caplog.at_level(logging.WARNING, target_win.log.name):
        assert len(list(target_win.tasks(group=True))) == 18
        assert "Invalid task file encountered:" in caplog.text
